import radical.utils as ru
from radical.entk.exceptions import *
from radical.entk.stage.stage import Stage
import threading
from radical.entk import states
from collections import Iterable

class Pipeline(object):

    """
    A pipeline represents a collection of objects that have a linear temporal execution order.
    In this case, a pipeline consists of multiple 'Stage' objects. Each ```Stage_i``` can execute only
    after all stages up to ```Stage_(i-1)``` have completed execution.

    """

    def __init__(self):

        self._uid       = ru.generate_id('radical.entk.pipeline')
        self._stages    = list()
        self._name      = str()

        self._state     = states.INITIAL

        # Keep track of states attained
        self._state_history = [states.INITIAL]

        # To keep track of current state
        self._stage_count = len(self._stages)
        self._cur_stage = 0

        # Lock around current stage
        self._lock = threading.Lock()

        # To keep track of termination of pipeline
        self._completed_flag = threading.Event()    

    # ------------------------------------------------------------------------------------------------------------------
    # Getter functions
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        """
        Name of the pipeline useful for bookkeeping and to refer to this pipeline while data staging

        :getter: Returns the name of the pipeline
        :setter: Assigns the name of the pipeline
        :type: String
        """
        return self._name
    
    @property
    def stages(self):

        """
        Stages of the list

        :getter: Returns the stages in the current Pipeline
        :setter: Assigns the stages to the current Pipeline
        :type: List
        """

        return self._stages

    @property
    def state(self):

        """
        Current state of the pipeline

        :getter: Returns the state of the current pipeline
        :type: String
        """

        return self._state


    @property
    def uid(self):

        """
        Unique ID of the current pipeline

        :getter: Returns the unique id of the current pipeline
        :type: String
        """
        return self._uid

    
    @property
    def _stage_lock(self):

        """
        Returns the lock over the current Pipeline

        :return: Lock object
        """

        return self._lock
    
    @property
    def completed(self):

        """
        Returns whether the Pipeline has completed

        :return: Boolean
        """
        return self._completed_flag.is_set()

    @property
    def current_stage(self):

        """
        Returns the current stage being executed

        :return: Integer
        """

        return self._cur_stage
    
    @property
    def state_history(self):

        """
        Returns a list of the states obtained in temporal order
        
        :return: list
        """

        return self._state_history

    # ------------------------------------------------------------------------------------------------------------------
    # Setter functions
    # ------------------------------------------------------------------------------------------------------------------

    @name.setter
    def name(self, value):
        if isinstance(value,str):
            self._name = value

        else:
            raise TypeError(expected_type=str, actual_type=type(value))

    @stages.setter
    def stages(self, val):
            
        self._stages = self._validate_stages(val)

        self._pass_uid()
        self._stage_count = len(self._stages)
        if self._cur_stage == 0:
            self._cur_stage = 1



    @state.setter
    def state(self, value):
        if isinstance(value,str):
            self._state = value
            self._state_history.append(value)
        else:
            raise TypeError(expected_type=str, actual_type=type(value)) 

    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def add_stages(self, val):

        """        
        Appends stages to the current Pipeline

        :argument: List of Stage objects
        """
        stages = self._validate_stages(val)

        stages = self._pass_uid(stages)
        self._stages.extend(stages)
        self._stage_count = len(self._stages)
        if self._cur_stage == 0:
            self._cur_stage = 1

  
    def to_dict(self):


        """
        Convert current Pipeline into a dictionary

        :return: python dictionary
        """

        pipeline_desc_as_dict = {

                                'uid': self._uid,
                                'name': self._name,
                                'state': self._state,
                                'state_history': self._state_history,
                                'completed': self._completed_flag.is_set()
                        }

        return pipeline_desc_as_dict


    def from_dict(self, d):

        """
        Create a Pipeline from a dictionary. The change is in inplace.

        :argument: python dictionary
        :return: None
        """


        if 'uid' in d:
            if isinstance(d['uid'], str) or isinstance(d['uid'], unicode):
                self._uid   = d['uid']
            else:
                raise TypeError(entity='uid', expected_type=str, actual_type=type(d['uid']))

        if 'name' in d:
            if isinstance(d['name'], str) or isinstance(d['name'], unicode):
                self._name = d['name']
            else:
                raise TypeError(entity='name', expected_type=str, actual_type=type(d['name']))

        if 'state' in d:
            if isinstance(d['state'], str) or isinstance(d['state'], unicode):
                self._state = d['state']
            else:
                raise TypeError(entity='state', expected_type=str, actual_type=type(d['state']))

        else:
            self._state = states.INITIAL


        if 'state_history' in d:
            if isinstance(d['state_history'], list):
                self._state_history = d['state_history']
            else:
                raise TypeError(entity='state_history', expected_type=list, actual_type=type(d['state_history']))

        if 'completed' in d:
            if isinstance(d['completed'], bool):
                if d['completed']:
                    self._completed_flag.set()
            else:
                raise TypeError(entity='completed', expected_type=bool, actual_type=type(d['completed']))


    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------

    def _pass_uid(self, stages=None):


        """
        Purpose: Pass current Pipeline's uid to all Stages. 

        Details: This allows us to trace the correct Pipeline if only the Stage is given.

        :argument: List of Stage objects (optional)
        :return: List of updated Stage objects
        """

        try:

            if stages is None:
                for stage in self._stages:
                    stage.parent_pipeline = self._uid
                    stage._pass_uid()
            else:
                for stage in stages:
                    stage.parent_pipeline = self._uid
                    stage._pass_uid()

                return stages

        except Exception, ex:
            raise Error(text=ex)


    def _increment_stage(self):

        """
        Purpose: Increment stage pointer. Also check if Pipeline has completed.
        """

        try:

            if self._cur_stage < self._stage_count:
                self._cur_stage+=1
            else:
                self._completed_flag.set()

        except Exception, ex:
            raise Error(text=ex)

    def _decrement_stage(self):

        """
        Purpose: Decrement stage pointer 
        """

        try:

            if self._cur_stage > 0:
                self._cur_stage -= 1
                self._completed_flag = threading.Event() # reset
            
        except Exception, ex:
            raise Error(text=ex)

    def _validate_stages(self, stages):

        """
        Purpose: Validate whether the 'stages' is of type list. Validate the description of each Stage.

        Details: This method is to be called before the resource request is placed. Currently, this method is called 
        when stages are added to the Pipeline.
        """
        if not stages:
            raise TypeError(expected_type=Stage, actual_type=type(stages))

        if not isinstance(stages, list):
            stages = [stages]

        for val in stages:
            if not isinstance(val, Stage):
                raise TypeError(expected_type=Stage, actual_type=type(val))

            val._validate()

        return stages


    def _validate(self):

        """
        Purpose: Validate that the state of the current Pipeline is 'DESCRIBED' (user has not meddled with it). Also 
        validate that the current Pipeline contains Stages.

        Details: This method is to be called before the resource request is placed. Currently, this method is called
        when the entire workflow is validated (in the AppManager).
        """

        if self._state is not states.INITIAL:
            
            raise ValueError(   object=self._uid, 
                                attribute='state', 
                                expected_value=states.INITIAL,
                                actual_value=self._state)

        if self._stages is None:

            raise MissingError( object=self._uid,
                                missing_attribute='stages')
    # ------------------------------------------------------------------------------------------------------------------


from sync_initiator import sync_with_master

def transition(obj, obj_type, new_state, channel, queue, profiler, logger):


    try:
        old_state = obj.state
        obj.state = new_state
                                
        if obj_type == 'Task':
            msg = obj.parent_stage
        elif obj_type == 'Stage':
            msg = obj.parent_pipeline
        else:
            msg=None

        profiler.prof('transition', 
                        uid=obj.uid, 
                        state=obj.state,
                        msg=msg)

        sync_with_master(   obj=obj, 
                            obj_type=obj_type,
                            channel = channel,
                            queue = queue, 
                            logger=logger, 
                            local_prof = profiler)

        logger.info('Transition of %s to new state %s successful'%(obj.uid, new_state))

    except Exception, ex:

        logger.error('Transition of %s to %s state failed, error: %s'%(obj.uid, new_state, ex))
        obj.state =  old_state
        sync_with_master(   obj=obj, 
                            obj_type=obj_type,
                            channel = channel,
                            queue= queue, 
                            logger=logger, 
                            local_prof = profiler)
        raise


__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
import radical.pilot as rp
import os


class ResourceManager(object):

    """
    A resource manager takes the responsibility of placing resource requests on different, possibly multiple,
    DCIs. Currently, the runtime system being used is RADICAL Pilot and hence the resource request is made via
    Pilot Jobs.

    :arguments: 
        :resource_desc: dictionary with details of the resource request + access credentials of the user 
        :example: resource_desc = {  'resource': 'xsede.stampede', 'walltime': 120, 'cores': 64, 'project: 'TG-abcxyz'}

    """

    def __init__(self, resource_desc):

        self._resource_desc = resource_desc

        self._session = None
        self._pmgr = None
        self._pilot = None
        self._resource = None
        self._walltime = None
        self._cores = None
        self._project = None
        self._access_schema = None
        self._queue = None

        self._mlab_url = os.environ.get('RADICAL_PILOT_DBURL', None)
        if not self._mlab_url:
            raise Error(text='RADICAL_PILOT_DBURL not defined. Please assign a valid mlab url')

        # Shared data list
        self._shared_data = list()

    # ------------------------------------------------------------------------------------------------------------------
    # Getter methods
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def pilot(self):
        """
        :getter: Return reference to the submitted Pilot
        """
        return self._pilot

    @property
    def session(self):
        """
        :getter: Return reference to the Radical Pilot session instance currently being used
        """
        return self._session

    @property
    def pmgr(self):
        """
        :getter: Return reference to the Radical Pilot manager currently being used
        """
        return self._pmgr

    @property
    def resource(self):
        """
        :getter: Return user specified resource name
        """
        return self._resource

    @property
    def walltime(self):
        """
        :getter: Return user specified walltime
        """
        return self._walltime

    @property
    def cores(self):
        """
        :getter: Return user specified number of cpus
        """
        return self._cores

    @property
    def project(self):
        """
        :getter: Return user specified project ID
        """
        return self._project

    @property
    def access_schema(self):
        """
        :getter: Return user specified access schema -- 'ssh' or 'gsissh' or None
        """
        return self._access_schema

    @property
    def queue(self):
        """
        :getter: Return user specified resource queue to be used
        """
        return self._queue

    @property
    def shared_data(self):
        """
        :getter: list of files to be staged to remote and that are common to multiple tasks
        """
        return self._shared_data

    @shared_data.setter
    def shared_data(self, data_list):

        if not isinstance(data_list, list):
            data_list = [data_list]

        for val in data_list:
            if not isinstance(val, str):
                raise TypeError(expected_type=str, actual_type=type(val))

        self._shared_data = data_list

    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def get_resource_allocation_state(self):
        """
        **Purpose**: Get the state of the resource allocation

        """

        if self._pilot:
            return self._pilot.state
        else:
            return None

    def completed_states(self):
        """
        **Purpose**: Test if a resouce allocation was submitted

        """

        return [rp.CANCELED, rp.FAILED, rp.DONE]

    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------

    def _validate_resource_desc(self, sid):
        """
        **Purpose**: Validate the resource description that was provided to ResourceManager

        :arguments: dictionary consisting of details of the resource request
        :return: boolean (valid/invalid)
        """

        self._uid = ru.generate_id('radical.entk.resource_manager.%(item_counter)04d', ru.ID_CUSTOM, namespace=sid)
        self._path = os.getcwd() + '/' + sid

        self._logger = ru.get_logger(self._uid, path=self._path)
        self._prof = ru.Profiler(name=self._uid, path=self._path)

        try:
            
            self._prof.prof('validating rdesc', uid=self._uid)

            self._logger.debug('Validating resource description')

            if not isinstance(self._resource_desc, dict):
                raise TypeError(expected_type=dict, actual_type=type(self._resource_desc))

            expected_keys = [   'resource',
                                'walltime',
                                'cores'
                            ]

            for key in expected_keys:
                if key not in self._resource_desc:
                    raise Error(text='Key %s does not exist in the resource description' % key)

            if not isinstance(self._resource_desc['resource'], str):
                raise TypeError(expected_type=str, actual_type=type(self._resource_desc['resource']))

            if not isinstance(self._resource_desc['walltime'], int):
                raise TypeError(expected_type=int, actual_type=type(self._resource_desc['walltime']))

            if not isinstance(self._resource_desc['cores'], int):
                raise TypeError(expected_type=int, actual_type=type(self._resource_desc['cores']))

            if 'project' in self._resource_desc:
                if (not isinstance(self._resource_desc['project'],str)) and (not self._resource_desc['project']):
                    raise TypeError(expected_type=str, actual_type=type(self._resource_desc['project']))            

            if 'access_schema' in self._resource_desc:
                if not isinstance(self._resource_desc['access_schema'], str):
                    raise TypeError(expected_type=str, actual_type=type(self._resource_desc['access_schema']))

            if 'queue' in self._resource_desc:
                if not isinstance(self._resource_desc['queue'], str):
                    raise TypeError(expected_type=str, actual_type=type(self._resource_desc['queue']))

            self._logger.info('Resource description validated')

            self._prof.prof('rdesc validated', uid=self._uid)

            return True

        except Exception, ex:
            self._logger.error('Failed to validate resource description, error: %s' % ex)
            raise

    def _populate(self):
        """
        **Purpose**: Populate the ResourceManager attributes with values provided in the resource description

        :arguments: valid dictionary consisting of details of the resource request
        """

        try:

            self._prof.prof('populating rmgr', uid=self._uid)

            self._logger.debug('Populating resource manager object')

            self._resource = self._resource_desc['resource']
            self._walltime = self._resource_desc['walltime']
            self._cores = self._resource_desc['cores']

            if 'project' in self._resource_desc:
                self._project = self._resource_desc['project']

            if 'access_schema' in self._resource_desc:
                self._access_schema = self._resource_desc['access_schema']

            if 'queue' in self._resource_desc:
                self._queue = self._resource_desc['queue']

            self._logger.debug('Resource manager population successful')

            self._prof.prof('rmgr populated', uid=self._uid)

        except Exception, ex:
            self._logger.error('Resource manager population unsuccessful')
            raise

    def _submit_resource_request(self):
        """
        **Purpose**: Function to initiate the resource request.

        Details: Currently, submits a Pilot job using the RADICAL Pilot runtime system.
        """

        try:

            self._prof.prof('creating rreq', uid=self._uid)

            def _pilot_state_cb(pilot, state):
                self._logger.info('Pilot %s state: %s' % (pilot.uid, state))

                if state == rp.FAILED:
                    self._logger.error('Pilot has failed')

                elif state == rp.DONE:
                    self._logger.error('Pilot has completed')

            self._session = rp.Session(dburl=self._mlab_url)

            self._pmgr = rp.PilotManager(session=self._session)
            self._pmgr.register_callback(_pilot_state_cb)

            pd_init = {
                'resource': self._resource,
                'runtime': self._walltime,
                'cores': self._cores,
                'project': self._project,
            }

            if self._access_schema:
                pd_init['access_schema'] = self._access_schema

            if self._queue:
                pd_init['queue'] = self._queue

            # Create Compute Pilot with validated resource description
            pdesc = rp.ComputePilotDescription(pd_init)

            self._prof.prof('rreq created', uid=self._uid)

            # Launch the pilot
            self._pilot = self._pmgr.submit_pilots(pdesc)

            shared_staging_directives = list()
            for data in self._shared_data:
                temp = {
                    'source': data,
                    'target': 'pilot:///' + os.path.basename(data)
                }
                shared_staging_directives.append(temp)

            self._pilot.stage_in(shared_staging_directives)

            self._prof.prof('rreq submitted', uid=self._uid)

            self._logger.info('Resource request submission successful.. waiting for pilot to go Active')

            # Wait for pilot to go active
            self._pilot.wait([rp.PMGR_ACTIVE, rp.FAILED])

            if self._pilot.state == rp.FAILED:
                raise Exception

            self._prof.prof('resource active', uid=self._uid)

            self._logger.info('Pilot is now active')

        except KeyboardInterrupt:

            if self._session:
                self._session.close()

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to exit callback thread gracefully...')
            raise KeyboardInterrupt

        except Exception, ex:
            self._logger.error('Resource request submission failed')
            raise

    def _cancel_resource_request(self, download_rp_profile=False):

        """
        **Purpose**: Cancel the resource request
        """

        try:

            self._prof.prof('canceling resource allocation', uid=self._uid)
            self._pilot.cancel()
            self._session.close(cleanup=False, download=download_rp_profile)
            self._prof.prof('resource allocation cancelled', uid=self._uid)

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to exit callback thread gracefully...')
            raise KeyboardInterrupt

        except Exception, ex:
            self._logger.error('Could not cancel resource request, error: %s' % ex)
            raise

    # ------------------------------------------------------------------------------------------------------------------



from radical.entk import Pipeline, Stage, Task, AppManager, ResourceManager
import os

# ------------------------------------------------------------------------------
# Set default verbosity

if os.environ.get('RADICAL_ENTK_VERBOSE') == None:
    os.environ['RADICAL_ENTK_VERBOSE'] = 'INFO'

if __name__ == '__main__':

    # Create a Pipeline object
    p = Pipeline()

    # Create a Stage object 
    s1 = Stage()

    # Create a Task object which creates a file named 'output.txt' of size 1 MB
    t1 = Task()    
    t1.executable = ['/bin/bash']   
    t1.arguments = ['-l', '-c', 'base64 /dev/urandom | head -c 1000000 > output.txt'] 

    # Add the Task to the Stage
    s1.add_tasks(t1)

    # Add Stage to the Pipeline
    p.add_stages(s1)


    # Create another Stage object
    s2 = Stage()
    s2.name = 'Stage 2'

    # Create a Task object
    t2 = Task()
    t2.executable = ['/bin/bash']    
    t2.arguments = ['-l', '-c', 'grep -o . output.txt | sort | uniq -c > ccount.txt']  
    # Copy data from the task in the first stage to the current task's location
    t2.copy_input_data = ['$Pipline_%s_Stage_%s_Task_%s/output.txt'%(p.uid, s1.uid, t1.uid)]
    # Download the output of the current task to the current location
    t2.download_output_data = ['ccount.txt']

    # Add the Task to the Stage
    s2.add_tasks(t2)

    # Add Stage to the Pipeline
    p.add_stages(s2)


    # Create a dictionary describe four mandatory keys:
    # resource, walltime, cores and project
    # resource is 'local.localhost' to execute locally
    res_dict = {

            'resource': 'local.localhost',
            'walltime': 10,
            'cores': 1,
            'project': '',
    }

    # Create Resource Manager object with the above resource description
    rman = ResourceManager(res_dict)

    # Create Application Manager
    appman = AppManager()

    # Assign resource manager to the Application Manager
    appman.resource_manager = rman

    # Assign the workflow as a set of Pipelines to the Application Manager
    appman.assign_workflow(set([p]))

    # Run the Application Manager
    appman.run()



import radical.pilot as rp
from radical.entk import Task
import radical.utils as ru
import traceback
from radical.entk.exceptions import *
import os

logger = ru.get_logger('radical.entk.task_processor')

def resolve_placeholders(path, placeholder_dict):

    """
    **Purpose**: Substitute placeholders in staging attributes of a Task with actual paths to the corresponding tasks.

    :arguments:
        :path: string describing the staging paths, possibly containing a placeholder
        :placeholder_dict: dictionary holding the values for placeholders

    """

    try:

        if isinstance(path, unicode):
            path = str(path)

        if not isinstance(path,str):
            raise TypeError(expected_type=str, actual_type=type(path))

        if '$' not in path:
            return path

        # Extract placeholder from path
        if len(path.split('>'))==1:
            placeholder = path.split('/')[0]
        else:
            if path.split('>')[0].strip().startswith('$'):
                placeholder = path.split('>')[0].strip().split('/')[0]
            else:
                placeholder = path.split('>')[1].strip().split('/')[0]

        # SHARED
        if placeholder == "$SHARED":
            return path.replace(placeholder, 'pilot://')

        # Expected placeholder format:
        # $Pipeline_{pipeline.uid}_Stage_{stage.uid}_Task_{task.uid}

        broken_placeholder = placeholder.split('/')[0].split('_')

        if not len(broken_placeholder) ==  6:
            raise ValueError(   
                                obj='placeholder',
                                attribute='length',
                                expected_value = '$Pipeline_{pipeline.uid}_Stage_{stage.uid}_Task_{task.uid} or $SHARED',
                                actual_value = broken_placeholder)

        pipeline_uid    = broken_placeholder[1]
        stage_uid       = broken_placeholder[3]
        task_uid        = broken_placeholder[5]

        resolved_placeholder = path.replace(placeholder,placeholder_dict[pipeline_uid][stage_uid][task_uid])

        return resolved_placeholder

    except Exception, ex:

        logger.error('Failed to resolve placeholder %s, error: %s'%(path, ex))
        raise 
    

def get_input_list_from_task(task, placeholder_dict):

    """
    Purpose: Parse a Task object to extract the files to be staged as the output. 

    Details: The extracted data is then converted into the appropriate RP directive depending on whether the data
    is to be copied/downloaded.

    :arguments: 
        :task: EnTK Task object
        :placeholder_dict: dictionary holding the values for placeholders

    :return: list of RP directives for the files that need to be staged out
    """


    try:

        if not isinstance(task, Task):
            raise TypeError(expected_type=Task, actual_type=type(task))

        input_data = []

        if task.link_input_data:

            for path in task.link_input_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                                'source': path.split('>')[0].strip(),                            
                                'target': path.split('>')[1].strip(),
                                'action': rp.LINK
                            }
                else:
                    temp = {
                                'source': path.split('>')[0].strip(),                            
                                'target': os.path.basename(path.split('>')[0].strip()),
                                'action': rp.LINK
                            }
                input_data.append(temp)

        if task.upload_input_data:

            for path in task.upload_input_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                                'source': path.split('>')[0].strip(),
                                'target': path.split('>')[1].strip()
                            }
                else:
                    temp = {
                                'source': path.split('>')[0].strip(),
                                'target': os.path.basename(path.split('>')[0].strip())
                            }
                input_data.append(temp)


        if task.copy_input_data:

            for path in task.copy_input_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                                'source': path.split('>')[0].strip(),
                                'target': path.split('>')[1].strip(),
                                'action': rp.COPY
                            }
                else:
                    temp = {
                                'source': path.split('>')[0].strip(),
                                'target': os.path.basename(path.split('>')[0].strip()),
                                'action': rp.COPY
                            }
                input_data.append(temp)

        return input_data

    except Exception, ex:

        logger.error('Failed to get input list of files from task, error: %s'%ex)
        raise 


def get_output_list_from_task(task, placeholder_dict):

    """
    Purpose: Parse a Task object to extract the files to be staged as the output. 
    
    Details: The extracted data is then converted into the appropriate RP directive depending on whether the data
    is to be copied/downloaded.
    
    :arguments: 
        :task: EnTK Task object
        :placeholder_dict: dictionary holding the values for placeholders

    :return: list of RP directives for the files that need to be staged out
    
    """


    try:

        if not isinstance(task, Task):
            raise TypeError(expected_type=Task, actual_type=type(task))


        output_data = []

        if task.copy_output_data:

            for path in task.copy_output_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                                'source': path.split('>')[0].strip(),
                                'target': path.split('>')[1].strip(),
                                'action': rp.COPY
                            }
                else:
                    temp = {
                                'source': path.split('>')[0].strip(),
                                'target': os.path.basename(path.split('>')[0].strip()),
                                'action': rp.COPY
                            }
                output_data.append(temp)


        if task.download_output_data:

            for path in task.download_output_data:

                path = resolve_placeholders(path, placeholder_dict)

                if len(path.split('>')) > 1:

                    temp = {
                                'source': path.split('>')[0].strip(),
                                'target': path.split('>')[1].strip()
                            }
                else:
                    temp = {
                                'source': path.split('>')[0].strip(),
                                'target': os.path.basename(path.split('>')[0].strip())
                            }
                output_data.append(temp)

        return output_data

    except Exception, ex:
        logger.error('Failed to get output list of files from task, error: %s'%ex)
        raise 


def create_cud_from_task(task, placeholder_dict, prof=None):

    """
    Purpose: Create a Compute Unit description based on the defined Task.

    :arguments: 
        :task: EnTK Task object
        :placeholder_dict: dictionary holding the values for placeholders

    :return: ComputeUnitDescription
    """

    try:
        
        logger.debug('Creating CU from Task %s'%(task.uid))

        if prof:
            prof.prof('cud from task - create', uid=task.uid)

        cud = rp.ComputeUnitDescription()
        cud.name        = '%s,%s,%s'%(task.uid, task.parent_stage, task.parent_pipeline)
        cud.pre_exec    = task.pre_exec
        cud.executable  = task.executable
        cud.arguments   = task.arguments
        cud.post_exec   = task.post_exec
        cud.cores       = task.cores
        cud.mpi         = task.mpi

        cud.input_staging   = get_input_list_from_task(task, placeholder_dict)
        cud.output_staging  = get_output_list_from_task(task, placeholder_dict)

        if prof:
            prof.prof('cud from task - done', uid=task.uid)

        logger.debug('CU %s created from Task %s'%(cud.name, task.uid))

        return cud

    except Exception, ex:        
        logger.error('CU creation failed, error: %s'%ex)
        raise


def create_task_from_cu(cu, prof=None):

    """
    Purpose: Create a Task based on the Compute Unit.

    Details: Currently, only the uid, parent_stage and parent_pipeline are retrieved. The exact initial Task (that was
    converted to a CUD) cannot be recovered as the RP API does not provide the same attributes for a CU as for a CUD. 
    Also, this is not required for the most part.

    TODO: Add exit code, stdout, stderr and path attributes to a Task. These can be extracted from a CU

    :arguments: 
        :cu: RP Compute Unit

    :return: Task
    """

    try:

        logger.debug('Create Task from CU %s'%cu.name)

        if prof:
            prof.prof('task from cu - create', uid=cu.name.split(',')[0].strip())

        task = Task(duplicate=True)
        task.uid                = cu.name.split(',')[0].strip()
        task.parent_stage      = cu.name.split(',')[1].strip()
        task.parent_pipeline   = cu.name.split(',')[2].strip()

        if cu.exit_code is not None:
            task.exit_code = cu.exit_code
        else:

            if cu.state == rp.DONE:
                task.exit_code = 0
            else:
                task.exit_code = 1

        task.path               =  ru.Url(cu.sandbox).path

        if prof:
            prof.prof('task from cu - done', uid=cu.name.split(',')[0].strip())

        logger.debug('Task %s created from CU %s'%(task.uid, cu.name))

        return task

    except Exception, ex:
        logger.error('Task creation from CU failed, error: %s'%ex)
        raise


import radical.utils as ru
from radical.entk.exceptions import *
from radical.entk.task.task import Task
from radical.entk import states
from collections import Iterable

class Stage(object):

    """
    A stage represents a collection of objects that have no relative order of execution. In this case, a
    stage consists of a set of 'Task' objects. All tasks of the same stage may execute concurrently.
    """

    def __init__(self, duplicate = False):

        if not duplicate:
            self._uid       = ru.generate_id('radical.entk.stage')
            
        self._tasks     = set()
        self._name      = str()

        self._state     = states.INITIAL

        # Keep track of states attained
        self._state_history = [states.INITIAL]

        # To change states
        self._task_count = len(self._tasks)

        # Pipeline this stage belongs to
        self._p_pipeline = None    

    # ------------------------------------------------------------------------------------------------------------------
    # Getter functions
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def name(self):

        """
        Name of the stage 

        :getter: Returns the name of the current stage
        :setter: Assigns the name of the current stage
        :type: String
        """
        return self._name
    
    @property
    def tasks(self):

        """
        Tasks of the stage

        :getter: Returns all the tasks of the current stage
        :setter: Assigns tasks to the current stage
        :type: set of Tasks
        """
        return self._tasks

    @property
    def state(self):

        """
        Current state of the stage

        :getter: Returns the state of the current stage
        :type: String
        """

        return self._state

    @property
    def parent_pipeline(self):

        """
        :getter: Returns the pipeline this stage belongs to
        :setter: Assigns the pipeline uid this stage belongs to
        """
        return self._p_pipeline

    @property
    def uid(self):

        """
        Unique ID of the current stage

        :getter: Returns the unique id of the current stage
        :type: String
        """

        return self._uid

    @property
    def state_history(self):

        """
        Returns a list of the states obtained in temporal order
        
        :return: list
        """

        return self._state_history


    # ------------------------------------------------------------------------------------------------------------------
    # Setter functions
    # ------------------------------------------------------------------------------------------------------------------

    @name.setter
    def name(self, value):
        if isinstance(value,str):
            self._name = value
        else:
            raise TypeError(expected_type=str, actual_type=type(value))
        

    @tasks.setter
    def tasks(self, val):        
        self._tasks = self._validate_tasks(val)
        self._task_count = len(self._tasks)
    
    @parent_pipeline.setter
    def parent_pipeline(self, value):
        if isinstance(value, str):
            self._p_pipeline = value
        else:
            raise TypeError(expected_type=str, actual_type=type(value))

    @state.setter
    def state(self, value):
        if isinstance(value,str):
            self._state = value
            self._state_history.append(value)
        else:
            raise TypeError(expected_type=str, actual_type=type(value))        
    

    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def add_tasks(self, val):

        """
        Adds tasks to the existing set of tasks of the Stage

        :argument: set of tasks
        """
        tasks =  self._validate_tasks(val)
        self._tasks.update(tasks)
        self._task_count = len(self._tasks)


    def to_dict(self):

        """
        Convert current Stage into a dictionary

        :return: python dictionary
        """

        stage_desc_as_dict = {

                                'uid': self._uid,
                                'name': self._name,
                                'state': self._state,
                                'state_history': self._state_history,
                                'parent_pipeline': self._p_pipeline
                        }

        return stage_desc_as_dict


    def from_dict(self, d):

        """
        Create a Stage from a dictionary. The change is in inplace.

        :argument: python dictionary
        :return: None
        """


        if 'uid' in d:
            if isinstance(d['uid'], str) or isinstance(d['uid'], unicode):
                self._uid   = d['uid']
            else:
                raise TypeError(entity='uid', expected_type=str, actual_type=type(d['uid']))

        if 'name' in d:
            if isinstance(d['name'], str) or isinstance(d['name'], unicode):
                self._name = d['name']
            else:
                raise TypeError(entity='name', expected_type=str, actual_type=type(d['name']))

        if 'state' in d:
            if isinstance(d['state'], str) or isinstance(d['state'], unicode):
                self._state = d['state']
            else:
                raise TypeError(entity='state', expected_type=str, actual_type=type(d['state']))

        else:
            self._state = states.INITIAL


        if 'state_history' in d:
            if isinstance(d['state_history'], list):
                self._state_history = d['state_history']
            else:
                raise TypeError(entity='state_history', expected_type=list, actual_type=type(d['state_history']))

        if 'parent_pipeline' in d:
            if isinstance(d['parent_pipeline'], str) or isinstance(d['parent_pipeline'], unicode):
                self._p_pipeline = d['parent_pipeline']
            else:
                raise TypeError(entity='parent_pipeline', expected_type=str, actual_type=type(d['parent_pipeline']))


    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------

    def _pass_uid(self, tasks=None):

        """
        Purpose: Assign the parent Stage and the parent Pipeline to all the tasks of the current stage. 

        Details: This lets us trace the Stage and Pipeline if only the Task is provided.

        :arguments: set of Tasks (optional)
        :return: list of updated Tasks
        """

        if tasks is None:
            for task in self._tasks:
                task.parent_stage = self._uid
                task.parent_pipeline = self._p_pipeline
        else:
            for task in tasks:
                task.parent_stage = self._uid
                task.parent_pipeline = self._p_pipeline


            return tasks

    def _set_tasks_state(self, value):

        """
        Purpose: Set state of all tasks of the current stage.

        :arguments: String
        """
        if value not in states.state_numbers.keys():
            raise ValueError(   obj=self._uid, 
                                attribute='set_tasks_state', 
                                expected_value = states.state_numbers.keys(), 
                                actual_value = value)

        for task in self._tasks:
            task.state = value


    def _check_stage_complete(self):

        """
        Purpose: Check if all tasks of the current stage have completed, i.e., are in either DONE or FAILED state.
        """

        try:

            for task in self._tasks:
                if task.state not in [states.DONE, states.FAILED]:
                    return False
                    
            return True

        except Exception, ex:

            print 'Task state evaluation failed'
            raise Error(text=ex)


    def _validate_tasks(self, tasks):

        """
        Purpose: Validate whether the 'tasks' is of type set. Validate the description of each Task.

        Details: This method is to be called before the resource request is placed. Currently, this method is called 
        when tasks are added to the stage.
        """        

        if not isinstance(tasks, set):

            if not isinstance(tasks, list):
                tasks = set([tasks])
            else:
                tasks = set(tasks)

        if not tasks:
            raise TypeError(expected_type=Task, actual_type=type(tasks))

        for t in tasks:

            if not isinstance(t, Task):
                raise TypeError(expected_type=Task, actual_type=type(t))

            t._validate()

        return tasks

    def _validate(self):

        """
        Purpose: Validate that the state of the current Stage is 'DESCRIBED' (user has not meddled with it). Also 
        validate that the current Stage contains Tasks

        Details: This method is to be called before the resource request is placed. Currently, this method is called
        when the parent Pipeline is validated.
        """

        if self._state is not states.INITIAL:
            
            raise ValueError(   obj=self._uid, 
                                attribute='state', 
                                expected_value=states.INITIAL,
                                actual_value=self._state)

        if not self._tasks:

            raise MissingError( obj=self._uid,
                                missing_attribute='tasks')

    # ------------------------------------------------------------------------------------------------------------------


import radical.utils as ru
from radical.entk.exceptions import *
from radical.entk import states

class Task(object):


    """
    A Task is an abstraction of a computational unit. In this case, a Task 
    consists of its executable along with its required software environment, 
    files to be staged as input and output.

    At the user level a Task is to be populated by assigning attributes. 
    Internally, an empty Task is created and population using the `from_dict`
    function. This is to avoid creating Tasks with new `uid` as tasks with new 
    `uid` offset the uid count file in radical.utils and can potentially affect
    the profiling if not taken care.
    """

    def __init__(self, duplicate = False):

        if not duplicate:
            self._uid       = ru.generate_id('radical.entk.task')

        self._name      = str()

        self._state     = states.INITIAL

        # Attributes necessary for execution
        self._pre_exec      = list()
        self._executable    = list()
        self._arguments     = list()
        self._post_exec     = list()
        self._cores         = 1
        self._mpi           = False

        # Data staging attributes
        self._upload_input_data     = list()
        self._copy_input_data       = list()
        self._link_input_data       = list()
        self._copy_output_data      = list()
        self._download_output_data  = list()

        self._path = None
        self._exit_code = None

        # Keep track of states attained
        self._state_history = [states.INITIAL]

        ## The following help in updation
        # Stage this task belongs to
        self._p_stage = None
        # Pipeline this task belongs to
        self._p_pipeline = None



    # ------------------------------------------------------------------------------------------------------------------
    # Getter functions
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def uid(self):

        """
        Unique ID of the current task

        :getter: Returns the unique id of the current task
        :type: String
        """

        if hasattr(self, '_uid'):
            return self._uid

    @property
    def name(self):

        """
        Name of the task

        :getter: Returns the name of the current task
        :setter: Assigns the name of the current task
        :type: String
        """

        return self._name
    
    @property
    def state(self):

        """
        Current state of the task

        :getter: Returns the state of the current task
        :type: String
        """

        return self._state
    
    @property
    def pre_exec(self):

        """
        List of commands to be executed prior to the executable

        :getter: return the list of commands
        :setter: assign the list of commands
        :arguments: list of strings
        """
        return self._pre_exec
    
    @property
    def executable(self):

        """
        A unix-based kernel to be executed

        :getter: returns the executable of the current task
        :setter: assigns the executable for the current task    
        :arguments: string
        """
        return self._executable
    
    @property
    def arguments(self):

        """
        List of arguments to be supplied to the executable

        :getter: returns the list of arguments of the current task
        :setter: assigns a list of arguments to the current task
        :arguments: list of strings
        """
        return self._arguments
    
    @property
    def post_exec(self):

        """
        List of commands to be executed post executable

        :getter: return the list of commands
        :setter: assign the list of commands
        :arguments: list of strings
        """

        return self._post_exec

    @property
    def cores(self):

        """
        List of commands to be executed post executable

        :getter: return the number of cores for this Task
        :setter: assign the number of cores for this Task
        """

        return self._cores


    @property
    def mpi(self):

        """
        List of commands to be executed post executable

        :getter: return MPI flag
        :setter: assign MPI flag
        """

        return self._mpi

    @property
    def upload_input_data(self):

        """
        List of files to be transferred from local machine to the location of the current task
        on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._upload_input_data
    
    @property
    def copy_input_data(self):

        """
        List of files to be copied from a location on the remote machine to the location of 
        current task on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._copy_input_data
    
    @property
    def link_input_data(self):

        """
        List of files to be linked from a location on the remote machine to the location of 
        current task on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._link_input_data
    
    @property
    def copy_output_data(self):

        """
        List of files to be copied from the location of the current task to another location
        on the remote machine

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """

        return self._copy_output_data
    
    @property
    def download_output_data(self):

        """
        List of files to be downloaded from the location of the current task to a location
        on the local machine.

        :getter: return the list of files
        :setter: assign the list of files
        :arguments: list of strings
        """
        return self._download_output_data

    @property
    def exit_code(self):

        """
        Get the exit code for DONE tasks. 0 for successful tasks, 1 for failed tasks.

        :getter: return the exit code of the current task
        """
        return self._exit_code


    @property
    def path(self):

        """
        Get the path of the task on the remote machine. Useful to reference files
        generated in the current task.

        :getter: return the path of the current task
        """

        return self._path

    @property
    def parent_stage(self):

        """
        :getter: Returns the stage this task belongs to
        :setter: Assigns the stage uid this task belongs to
        """
        return self._p_stage
    
    @property
    def parent_pipeline(self):

        """
        :getter: Returns the pipeline this task belongs to
        :setter: Assigns the pipeline uid this task belongs to
        """

        return self._p_pipeline    

    @property
    def state_history(self):

        """
        Returns a list of the states obtained in temporal order
        
        :return: list
        """

        return self._state_history


    # ------------------------------------------------------------------------------------------------------------------
    # Setter functions
    # ------------------------------------------------------------------------------------------------------------------

    @uid.setter
    def uid(self, val):
        if isinstance(val, str):
            self._uid = val
        else:
            raise TypeError(expected_type=str, actual_type=type(val))

    @name.setter
    def name(self, val):
        if isinstance(val,str):
            self._name = val
        else:
            raise TypeError(expected_type=str, actual_type=type(val))

    @state.setter
    def state(self, val):
        if isinstance(val,str):
            self._state = val
            self._state_history.append(val)
        else:
            raise TypeError(expected_type=str, actual_type=type(val))

    @pre_exec.setter
    def pre_exec(self, val):
        if isinstance(val, list):
            self._pre_exec = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))


    @executable.setter
    def executable(self, val):
        if isinstance(val, list):
            self._executable = val
        elif isinstance(val, str):
            self._executable = [val]
        else:
            raise TypeError(expected_type='list or str', actual_type=type(val))

    @arguments.setter
    def arguments(self, val):
        if isinstance(val, list):
            self._arguments = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))


    @post_exec.setter
    def post_exec(self, val):
        if isinstance(val, list):
            self._post_exec = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))


    @cores.setter
    def cores(self, val):
        if isinstance(val, int):
            if val > 0:
                self._cores = val
            else:
                raise ValueError(obj=self._uid, attribute='cores', expected_value='int > 0', actual_value=val)
        else:
            raise TypeError(expected_type=int, actual_type=type(val))


    @mpi.setter
    def mpi(self, val):
        if isinstance(val, bool):
            self._mpi = val
        else:
            raise TypeError(expected_type=bool, actual_type=type(val))


    @upload_input_data.setter
    def upload_input_data(self, val):
        if isinstance(val, list):
            self._upload_input_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @copy_input_data.setter
    def copy_input_data(self, val):
        if isinstance(val, list):
            self._copy_input_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @link_input_data.setter
    def link_input_data(self, val):
        if isinstance(val, list):
            self._link_input_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @copy_output_data.setter
    def copy_output_data(self, val):
        if isinstance(val, list):
            self._copy_output_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @download_output_data.setter
    def download_output_data(self, val):
        if isinstance(val, list):
            self._download_output_data = val
        else:
            raise TypeError(expected_type=list, actual_type=type(val))

    @exit_code.setter
    def exit_code(self, val):
        if isinstance(val, int):
            self._exit_code = val
        else:
            raise TypeError(entity='exit_code', expected_type=int, actual_type=type(val))


    @path.setter
    def path(self, val):
        if isinstance(val, str):
            self._path = val
        else:
            raise TypeError(entity='path', expected_type=str, actual_type=type(val))

    @parent_stage.setter
    def parent_stage(self, val):
        if isinstance(val,str):
            self._p_stage = val
        else:
            raise TypeError(expected_type=str, actual_type=type(val))

    @parent_pipeline.setter
    def parent_pipeline(self, val):
        if isinstance(val,str):
            self._p_pipeline = val
        else:
            raise TypeError(expected_type=str, actual_type=type(val))



    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def to_dict(self):

        """
        Convert current Task into a dictionary

        :return: python dictionary
        """

        task_desc_as_dict = {
                        'uid': self._uid,
                        'name': self._name,
                        'state': self._state,
                        'state_history': self._state_history,

                        'pre_exec': self._pre_exec,
                        'executable': self._executable,
                        'arguments': self._arguments,
                        'post_exec': self._post_exec,
                        'cores': self._cores,
                        'mpi': self._mpi,

                        'upload_input_data': self._upload_input_data,
                        'copy_input_data': self._copy_input_data,
                        'link_input_data': self._link_input_data,
                        'copy_output_data': self._copy_output_data,
                        'download_output_data': self._download_output_data,

                        'exit_code': self._exit_code,
                        'path': self._path,

                        'parent_stage': self._p_stage,
                        'parent_pipeline': self._p_pipeline,
                    }

        return task_desc_as_dict
        


    def from_dict(self, d):

        """
        Create a Task from a dictionary. The change is in inplace.

        :argument: python dictionary
        :return: None
        """

        if 'uid' in d:
            if isinstance(d['uid'], str) or isinstance(d['uid'], unicode):
                self._uid   = d['uid']
            else:
                raise TypeError(entity='uid', expected_type=str, actual_type=type(d['uid']))

        if 'name' in d:
            if isinstance(d['name'], str) or isinstance(d['name'], unicode):
                self._name = d['name']
            else:
                raise TypeError(entity='name', expected_type=str, actual_type=type(d['name']))

        if 'state' in d:
            if isinstance(d['state'], str) or isinstance(d['state'], unicode):
                self._state = d['state']
            else:
                raise TypeError(entity='state', expected_type=str, actual_type=type(d['state']))
        else:
            self._state = states.INITIAL


        if 'state_history' in d:
            if isinstance(d['state_history'], list):
                self._state_history = d['state_history']
            else:
                raise TypeError(entity='state_history', expected_type=list, actual_type=type(d['state_history']))


        if 'pre_exec' in d:
            if isinstance(d['pre_exec'], list):
                self._pre_exec = d['pre_exec']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['pre_exec']))

        if 'executable' in d:
            if isinstance(d['executable'], list):
                self._executable = d['executable']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['executable']))

        if 'arguments' in d:
            if isinstance(d['arguments'], list):
                self._arguments = d['arguments']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['arguments']))

        if 'post_exec' in d:
            if isinstance(d['post_exec'], list):
                self._post_exec = d['post_exec']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['post_exec']))

        if 'cores' in d:
            if isinstance(d['cores'], int):
                self._cores = d['cores']
            else:
                raise TypeError(expected_type=int, actual_type=type(d['cores']))


        if 'mpi' in d:
            if isinstance(d['mpi'], bool):
                self._mpi = d['mpi']
            else:
                raise TypeError(expected_type=bool, actual_type=type(d['mpi']))
            
        if 'upload_input_data' in d:
            if isinstance(d['upload_input_data'], list):
                self._upload_input_data = d['upload_input_data']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['upload_input_data']))

        if 'copy_input_data' in d:
            if isinstance(d['copy_input_data'], list):
                self._copy_input_data = d['copy_input_data']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['copy_input_data']))
            

        if 'link_input_data' in d:
            if isinstance(d['link_input_data'], list):
                self._link_input_data = d['link_input_data']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['link_input_data']))

        if 'copy_output_data' in d:
            if isinstance(d['copy_output_data'], list):
                self._copy_output_data = d['copy_output_data']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['copy_output_data']))

        if 'download_output_data' in d:
            if isinstance(d['download_output_data'], list):
                self._download_output_data = d['download_output_data']
            else:
                raise TypeError(expected_type=list, actual_type=type(d['download_output_data']))

        if 'exit_code' in d:
            if d['exit_code']:
                if isinstance(d['exit_code'], int):
                    self._exit_code = d['exit_code']
                else:
                    raise TypeError(entity='exit_code', expected_type=int, actual_type=type(d['exit_code']))

        if 'path' in d:
            if d['path']:
                if isinstance(d['path'], str) or isinstance(d['path'], unicode):
                    self._path = d['path']
                else:
                    raise TypeError(entity='path',expected_type=str, actual_type=type(d['path']))

        if 'parent_stage' in d:
            if isinstance(d['parent_stage'], str) or isinstance(d['parent_stage'], unicode):
                self._p_stage = d['parent_stage']
            else:
                raise TypeError(entity='parent_stage', expected_type=str, actual_type=type(d['parent_stage']))            

        if 'parent_pipeline' in d:
            if isinstance(d['parent_pipeline'], str) or isinstance(d['parent_pipeline'], unicode):
                self._p_pipeline = d['parent_pipeline']
            else:
                raise TypeError(entity='parent_pipeline', expected_type=str, actual_type=type(d['parent_pipeline']))

    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------

    def _validate(self):

        """
        Purpose: Validate that the state of the task is 'DESCRIBED' and that an executable has been specified for the 
        task. 
        """

        if self._state is not states.INITIAL:
            raise ValueError(   obj=self._uid, 
                                attribute='state', 
                                expected_value=states.INITIAL,
                                actual_value=self._state)

        if not self._executable:
            raise MissingError( obj=self._uid,
                                missing_attribute='executable')
    # ------------------------------------------------------------------------------------------------------------------


__copyright__   = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__      = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__     = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
import threading
from multiprocessing import Process, Event
import Queue
from radical.entk import states, Task
import time
import json
import pika
import traceback
import os

slow_run = os.environ.get('RADICAL_ENTK_SLOW',False)

# TEMPORARY FILE TILL AN EXECUTION PLUGIN IS IN PLACE

class Helper(object):

    def __init__(self, pending_queue, completed_queue, mq_hostname, resource_desc):

        self._uid           = ru.generate_id('radical.entk.helper')
        self._logger        = ru.get_logger('radical.entk.helper')

        self._pending_queue = pending_queue
        self._completed_queue = completed_queue
        self._mq_hostname = mq_hostname

        self._helper_process = None

        self._resource_desc = resource_desc

        self._logger.info('Created helper object: %s'%self._uid)


    def start_helper(self):

        # This method starts the extractor function in a separate thread        
        
        if not self._helper_process:

            try:

                # Attempt to start RP pilot on 


            try:
                self._helper_process = Process(target=self.helper, name='helper')
                self._helper_terminate = Event()
                self._logger.info('Starting helper process')
                self._helper_process.start()                

                return True

            except Exception, ex:

                self.end_helper()
                self._logger.error('Helper process not started')
                raise

                

        else:
            self._logger.info('Helper process already running')
            raise


    def end_helper(self):

        # Set termination flag
        try:
            if not self._helper_terminate.is_set():
                self._helper_terminate.set()

            self._helper_process.join()
            self._logger.info('Helper thread closed')

            return True

        except Exception, ex:
            self._logger.error('Could not terminate helper process')
            raise



    def helper(self):

        # This function extracts currently tasks from the pending_queue
        # and pushes it to the executed_queue. Thus mimicking an execution plugin

        try:

            self._logger.info('Helper process started')

            # Thread should run till terminate condtion is encountered
            mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname))
            mq_channel = mq_connection.channel()

            while not self._helper_terminate.is_set():

                try:

                    method_frame, header_frame, body = mq_channel.basic_get(queue=self._pending_queue[0])

                    if body:


                        try:
                            task = Task()
                            task.load_from_dict(json.loads(body))

                            task.state = states.DONE

                            task_as_dict = json.dumps(task.to_dict())

                            self._logger.debug('Got task %s from pending_queue %s'%(task.uid, self._pending_queue[0]))

                            mq_channel.basic_publish( exchange='fork',
                                                            routing_key='',
                                                            body=task_as_dict
                                                            #properties=pika.BasicProperties(
                                                                # make message persistent
                                                            #    delivery_mode = 2, 
                                                            #)
                                                        )                    
                        
                            self._logger.debug('Pushed task %s with state %s to completed queue %s and synchronizerq'%(
                                                                                    task.uid, 
                                                                                    task.state,
                                                                                    self._completed_queue[0])
                                                                                )

                            mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                        except Exception, ex:

                            # Rolling back queue and task status
                            self._logger.error('Error while pushing task to completed queue, rolling back: %s'%ex)
                            raise UnknownError(text=ex)
                
                        if slow_run:
                            time.sleep(1)

                


                except Exception, ex:

                    self._logger.error('Error getting messages from pending queue: %s'%ex)
                    raise UnknownError(text=ex) 


        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), '+
                                'trying to cancel enqueuer thread gracefully...')
            raise KeyboardInterrupt


        except Exception, ex:

            self._logger.error('Unknown error in helper process: %s'%ex)
            print traceback.format_exc()
            raise UnknownError(text=ex) 


    def check_alive(self):

        return self._helper_process.is_alive()


import uuid
import json
import pika

def sync_with_master(obj, obj_type, channel, queue, logger, local_prof):

    object_as_dict = {'object': obj.to_dict()}
    if obj_type == 'Task': 
        object_as_dict['type'] = 'Task'

    elif obj_type == 'Stage':
        object_as_dict['type'] = 'Stage'

    elif obj_type == 'Pipeline':
        object_as_dict['type'] = 'Pipeline'

    corr_id = str(uuid.uuid4())

    logger.debug('Attempting to sync %s with state %s with AppManager'%(obj.uid, obj.state))
    channel.basic_publish(
                            exchange='',
                            routing_key=queue,
                            body=json.dumps(object_as_dict),
                            properties=pika.BasicProperties(
                                        correlation_id = corr_id
                                        )
                        )

    if obj_type == 'Task':
        local_prof.prof('publishing obj with state %s for sync'%obj.state, uid=obj.uid, msg=obj.parent_stage)
    elif obj_type == 'Stage':
        local_prof.prof('publishing obj with state %s for sync'%obj.state, uid=obj.uid, msg=obj.parent_pipeline)
    else:
        local_prof.prof('publishing obj with state %s for sync'%obj.state, uid=obj.uid)


    sid = '-'.join(queue.split('-')[:-3])
    qname = queue.split('-')[-3:]
    qname.reverse()
    reply_queue = '-'.join(qname)
    reply_queue = sid + '-' + reply_queue
                
    while True:
        #self._logger.info('waiting for ack')        

        method_frame, props, body = channel.basic_get(queue=reply_queue)

        if body:
            if corr_id == props.correlation_id:

                #print 'acknowledged: ', obj.uid, obj.state
                if obj_type == 'Task':
                    local_prof.prof('obj with state %s synchronized'%obj.state, uid=obj.uid, msg=obj.parent_stage)
                elif obj_type == 'Stage':
                    local_prof.prof('obj with state %s synchronized'%obj.state, uid=obj.uid, msg=obj.parent_pipeline)
                else:
                    local_prof.prof('obj with state %s synchronized'%obj.state, uid=obj.uid)
                            
                logger.debug('%s with state %s synced with AppManager'%(obj.uid, obj.state))

                channel.basic_ack(delivery_tag = method_frame.delivery_tag)

                break


__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import pandas as pd
import argparse
import numpy as np
import pprint
import os

class Profiler(object):

    """
    The Profiler object to read EnTK profiles are produce data frames and 
    durations based on states and events within EnTK.

    :Arguments:
        :src: full path to the folder containing the EnTK profiles        
    """

    def __init__(self, src=None):

        if not src:
            src = './'


        self._src = src

        self._states = dict()
        self._events = dict()

        self._legacy = os.environ.get("RADICAL_ANALYTICS_LEGACY_PROFILES", False)
        if self._legacy:
            self._cols = ['timestamp', 'thread/proc', 'uid', 'state', 'event', 'msg']
        else:
            self._cols = ['timestamp', 'event', 'comp', 'tid', 'uid', 'state', 'msg']

        self._initialize()

    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def get(self, uid=None, state=None):

        """
        Get a specific object as specified by the 'uid' argument or
        get a list of objects with a specific state as specified by the 'state'
        argument

        :Arguments:
            :uid: Uid of the object whose details are required
            :state: Details of all objects that have the specified state
        """

        if uid and not state:

            data = dict()            

            obj_type = uid.split('.')[0]

            for state, object_collection in self._states_dict.iteritems():
                
                for obj, timestamp in object_collection[obj_type].iteritems():

                    if obj == uid:

                        if 'uid' not in data:
                            data['uid'] = uid

                        data[state] = timestamp

            return data

        elif state and not uid:

            data = dict()

            for object_collection, objs in self._states_dict[state].iteritems():

                for obj, timestamp in objs.iteritems():

                    if 'state' not in data:
                        data['state'] = state

                    data[obj] = timestamp

            return data

        else:

            data = dict()
            obj_type = uid.split('.')[0]

            data['uid'] = uid
            data[state] = self._states_dict[state][obj_type][uid]

            return data


    def duration(self, objects, states=None, events=None):

        """
        Get the duration the list of objects spent between the list of states or 
        events specified

        :Arguments:
            :objects: List of uids of objects across which the duration is to be computed
            :states: The states between which the duration is to be computed
                syntax: [start_state, end_state]
            :events: The events between which the duration is to be computed
                syntax: [start_event, end_event]
        """

        if not isinstance(states, list) and (states is not None):
            states = [states]

        if not isinstance(events, list) and (events is not None):
            events = [events]

        if not isinstance(objects, list):
            objects =[objects]

        if len(objects)==1:
            objects = [objects[0], objects[0]]

        #pprint.pprint(self._states_dict)
        #pprint.pprint(self._events_dict)

        # Check if objects are in accepted format

        # extract info about required objects

        extracted_dict = {}

        '''
        {
            object0: 
                    { 
                        state0: time, 
                        state1: time
                    },
            object1: 
                    { 
                        state0: time, 
                        state1: time
                    },
        }

        '''

        if (states is not None) and events is None:

            for obj in objects:

                extracted_dict[obj] = {}
                extracted_dict[obj][states[0]] = self._states_dict[states[0]][obj.split('.')[2].strip()][obj]
                extracted_dict[obj][states[1]] = self._states_dict[states[1]][obj.split('.')[2].strip()][obj]

            #pprint.pprint(extracted_dict)

            return self._get_Toverlap(extracted_dict, states[0], states[1])


        elif (states is None) and (events is not None):

            for obj in objects:

                extracted_dict[obj] = {}
                extracted_dict[obj][events[0]] = self._events_dict[events[0]][obj.split('.')[2].strip()][obj]
                extracted_dict[obj][events[1]] = self._events_dict[events[1]][obj.split('.')[2].strip()][obj]            

            #pprint.pprint(extracted_dict)

            return self._get_Toverlap(extracted_dict, events[0], events[1])


        elif states is not None and events is not None:

            t1 = None
            t2 = None

            for obj in objects:

                
                if obj.split('.')[0].strip() in ['task', 'stage', 'pipeline']:
                    t1 = float(self._states_dict[states[0]][obj.split('.')[2].strip()][obj])
                else:
                    t2 = float(self._events_dict[events[0]][obj.split('.')[2].strip()][obj])

            if t1>t2:
                return float(t1) - float(t2)
            else:
                return float(t2) - float(t1)


    def list_all_stateful_objects(self):

        """
        List all objects and the details as stored in the profiler
        """

        return self._state_objs

    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------

    def _initialize(self):
      
        self._get_resource_manager_details()
        self._get_app_manager_details()
        self._get_wfp_details()
        self._get_task_manager_details()

        self._get_state_details()
        self._get_event_details()

        self._get_states_dict()
        self._get_events_dict()


    def _get_resource_manager_details(self):
    
        self._rman_df = pd.read_csv('%s/radical.entk.resource_manager.0000.prof'%self._src,skiprows=[0,1], 
                            names=self._cols)


    def _get_app_manager_details(self):
    
        self._aman_df = pd.read_csv('%s/radical.entk.appmanager.0000.prof'%self._src,skiprows=[0,1], 
                            names=self._cols)
    
    
    def _get_wfp_details(self):
    
        df_obj = pd.read_csv('%s/radical.entk.wfprocessor.0000-obj.prof'%self._src,skiprows=[0,1], 
                            names=self._cols)
    
        df_proc = pd.read_csv('%s/radical.entk.wfprocessor.0000-proc.prof'%self._src,skiprows=[0,1], 
                            names=self._cols)
    
    
        self._wfp_df = pd.concat([df_obj,df_proc]).sort_values(by='timestamp')
    
    
    def _get_task_manager_details(self):
    
        df_obj = pd.read_csv('%s/radical.entk.task_manager.0000-obj.prof'%self._src,skiprows=[0,1], 
                            names = self._cols)
    
        df_proc = pd.read_csv('%s/radical.entk.task_manager.0000-proc.prof'%self._src,skiprows=[0,1], 
                            names = self._cols)
    
    
        self._tman_df = pd.concat([df_obj,df_proc]).sort_values(by='timestamp')
    
    
    def _get_state_details(self):
    
        state_df  = pd.DataFrame(columns=self._cols)
    
        state_df = state_df.append(self._wfp_df[pd.notnull(self._wfp_df['state'])]).sort_values(by='timestamp')
        state_df = state_df.append(self._tman_df[pd.notnull(self._tman_df['state'])]).sort_values(by='timestamp')
    
        self._states_df = state_df.reset_index().drop(['index'], axis=1)
    
    
    
    def _get_event_details(self):
    
        event_df  = pd.DataFrame(columns=self._cols)    

        event_df = event_df.append(self._rman_df[pd.isnull(self._rman_df['state'])]).sort_values(by='timestamp')
        event_df = event_df.append(self._aman_df[pd.isnull(self._aman_df['state'])]).sort_values(by='timestamp')
        event_df = event_df.append(self._wfp_df[pd.isnull(self._wfp_df['state'])]).sort_values(by='timestamp')
        event_df = event_df.append(self._tman_df[pd.isnull(self._tman_df['state'])]).sort_values(by='timestamp')
    
        self._events_df = event_df.reset_index().drop(['index'], axis=1)
    


    def _get_states_dict(self):

        self._states_dict = dict()
        self._state_objs = dict()

        for row in self._states_df.iterrows():

            row = row[1]

            if row['state'] not in self._states_dict:
                self._states_dict[row['state']] = dict()

            obj_type = row['uid'].split('.')[2]
            #obj_name = ''.join([row['uid'].split('.')[2],'.',row['uid'].split('.')[3]])
            obj_name = row['uid']

            if obj_type not in self._states_dict[row['state']]:
                self._states_dict[row['state']][obj_type] = dict()

            self._states_dict[row['state']][obj_type][obj_name] = row['timestamp']


        for row in self._states_df.iterrows():

            row = row[1]

            if 'pipeline' in row['uid']:
                if row['uid'] not in self._state_objs:
                    self._state_objs[row['uid']] = dict()

            elif 'stage' in row['uid']:
                if row['uid'] not in self._state_objs[row['msg']]:
                    self._state_objs[row['msg']][row['uid']] = list()

            elif 'task' in row['uid']:
                for pipeline, stages in self._state_objs.iteritems():
                    for stage, tasks in stages.iteritems():
                        if row['msg'] == stage:
                            if row['uid'] not in tasks:
                                tasks.append(row['uid'])

    def _get_events_dict(self):

        self._events_dict = dict()

        for row in self._events_df.iterrows():

            row = row[1]

            if row['event'] not in self._events_dict:
                self._events_dict[row['event']] = dict()



            if pd.notnull(row['uid']):
                obj_type = row['uid'].split('.')[2]
                #obj_name = ''.join([row['uid'].split('.')[2],'.',row['uid'].split('.')[3]])
                obj_name = row['uid']

                if obj_type not in self._events_dict[row['event']]:
                    self._events_dict[row['event']][obj_type] = dict()

                self._events_dict[row['event']][obj_type][obj_name] = row['timestamp']


        #pprint.pprint(self._events_dict)


    def _get_Toverlap(self, d, start_state, stop_state):
        '''
        Helper function to create the list of lists from which to calculate the
        overlap of the elements of a DataFrame between the two boundaries passed as
        arguments.
        '''

        overlap = 0
        ranges = []

        for obj, states in d.iteritems():
            #print states
            ranges.append([states[start_state], states[stop_state]])



        for crange in self._collapse_ranges(ranges):
            overlap += crange[1] - crange[0]

        return overlap

    def _collapse_ranges(self, ranges):
        """
        given be a set of ranges (as a set of pairs of floats [start, end] with
        'start <= end'. This algorithm will then collapse that set into the
        smallest possible set of ranges which cover the same, but not more nor
        less, of the domain (floats).
    
        We first sort the ranges by their starting point. We then start with the
        range with the smallest starting point [start_1, end_1], and compare to the
        next following range [start_2, end_2], where we now know that start_1 <=
        start_2. We have now two cases:
    
        a) when start_2 <= end_1, then the ranges overlap, and we collapse them
        into range_1: range_1 = [start_1, max[end_1, end_2]
    
        b) when start_2 > end_2, then ranges don't overlap. Importantly, none of
        the other later ranges can ever overlap range_1. So we move range_1 to
        the set of final ranges, and restart the algorithm with range_2 being
        the smallest one.
    
        Termination condition is if only one range is left -- it is also moved to
        the list of final ranges then, and that list is returned.
        """

        final = []

        # sort ranges into a copy list
        _ranges = sorted (ranges, key=lambda x: x[0])

        START = 0
        END = 1

        base = _ranges[0] # smallest range

        for _range in _ranges[1:]:

            if _range[START] <= base[END]:

                # ranges overlap -- extend the base
                base[END] = max(base[END], _range[END])

            else:

                # ranges don't overlap -- move base to final, and current _range
                # becomes the new base
                final.append(base)
                base = _range

        # termination: push last base to final
        final.append(base)

        return final


    


__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
from multiprocessing import Process, Event
from radical.entk import states, Pipeline, Task
from radical.entk.utils.init_transition import transition
import time
from time import sleep
import json
import threading
import pika
import traceback
import os
import uuid


class WFprocessor(object):

    """
    An WFProcessor (workflow processor) takes the responsibility of dispatching tasks from the various pipelines of the
    workflow according to their relative order to the TaskManager. All state updates are communicated to the AppManager.

    :Arguments:
        :workflow: COPY of the entire workflow existing in the AppManager 
        :pending_qs: number of queues to hold pending tasks
        :completed_qs: number of queues to hold completed tasks
        :mq_hostname: hostname where the RabbitMQ is live
        :port: port at which rabbitmq can be accessed
    """

    def __init__(self,
                 sid,
                 workflow,
                 pending_queue,
                 completed_queue,
                 mq_hostname,
                 port,
                 resubmit_failed):


        if isinstance(sid, str):
            self._sid = sid
        else:
            raise TypeError(expected_type=str, actual_type=type(sid))

        self._uid = ru.generate_id('radical.entk.wfprocessor.%(item_counter)04d', ru.ID_CUSTOM, namespace=self._sid)
        self._path = os.getcwd() + '/' + self._sid

        self._logger = ru.get_logger(self._uid, path=self._path)
        self._prof = ru.Profiler(name=self._uid + '-obj', path=self._path)

        self._prof.prof('create wfp obj', uid=self._uid)

        self._workflow = workflow
        self._validate_workflow()


        if not isinstance(pending_queue, list):
            raise TypeError(expected_type=list, actual_type=type(pending_queue))
        self._pending_queue = pending_queue

        if not isinstance(completed_queue, list):
            raise TypeError(expected_type=list, actual_type=type(completed_queue))
        self._completed_queue = completed_queue

        if not isinstance(mq_hostname, str) and not isinstance(mq_hostname, unicode):
            raise TypeError(expected_type=str, actual_type=type(mq_hostname))
        self._mq_hostname = mq_hostname

        if not isinstance(port, int):
            raise TypeError(expected_type=int, actual_type=type(port))
        self._port = port

        if not isinstance(resubmit_failed, bool):
            raise TypeError(expected_type=bool, actual_type=type(resubmit_failed))
        self._resubmit_failed = resubmit_failed

        self._wfp_process = None
        
        self._logger.info('Created WFProcessor object: %s' % self._uid)

        self._prof.prof('wfp obj created', uid=self._uid)

        
    # ------------------------------------------------------------------------------------------------------------------
    # Getter
    # ------------------------------------------------------------------------------------------------------------------
    @property
    def workflow(self):
        return self._workflow

    # ------------------------------------------------------------------------------------------------------------------
    # Private Methods
    # ------------------------------------------------------------------------------------------------------------------

    def _validate_workflow(self):
        """
        **Purpose**: Validate whether the workflow consists of a set of Pipelines and validate each Pipeline. 

        Details: Tasks are validated when being added to Stage. Stages are validated when being added to Pipelines. Only
        Pipelines themselves remain to be validated before execution.
        """

        try:

            self._prof.prof('validating workflow', uid=self._uid)

            if not isinstance(self._workflow, set):

                if not isinstance(self._workflow, list):
                    self._workflow = set([self._workflow])
                else:
                    self._workflow = set(self._workflow)

            for item in self._workflow:
                if not isinstance(item, Pipeline):
                    self._logger.info('workflow type incorrect')
                    raise TypeError(expected_type=['Pipeline', 'set of Pipeline'],
                                    actual_type=type(item))

                item._validate()

            self._prof.prof('workflow validated', uid=self._uid)

        except Exception, ex:

            self._logger.error('Fatal error while adding workflow to appmanager: %s' % ex)
            raise

    #

    def _wfp(self):
        """
        **Purpose**: This is the function executed in the wfp process. The function is used to simply create
        and spawn two threads: enqueue, dequeue. The enqueue thread pushes ready tasks to the queues in the pending_q slow
        list whereas the dequeue thread pulls completed tasks from the queues in the completed_q. This function is also
        responsible for the termination of these threads and hence blocking.
        """

        try:

            local_prof = ru.Profiler(name=self._uid + '-proc', path=self._path)

            local_prof.prof('wfp process started', uid=self._uid)

            self._logger.info('WFprocessor started')

            # Process should run till terminate condtion is encountered
            while (not self._wfp_terminate.is_set()):

                try:

                    # Start dequeue thread
                    if (not self._dequeue_thread) or (not self._dequeue_thread.is_alive()):

                        local_prof.prof('creating dequeue-thread', uid=self._uid)
                        self._dequeue_thread = threading.Thread(
                            target=self._dequeue, args=(local_prof,), name='dequeue-thread')

                        self._logger.info('Starting dequeue-thread')
                        local_prof.prof('starting dequeue-thread', uid=self._uid)
                        self._dequeue_thread.start()

                    # Start enqueue thread
                    if (not self._enqueue_thread) or (not self._enqueue_thread.is_alive()):

                        local_prof.prof('creating enqueue-thread', uid=self._uid)
                        self._enqueue_thread = threading.Thread(
                            target=self._enqueue, args=(local_prof,), name='enqueue-thread')

                        self._logger.info('Starting enqueue-thread')
                        local_prof.prof('starting enqueue-thread', uid=self._uid)
                        self._enqueue_thread.start()

                except Exception, ex:
                    self._logger.error('WFProcessor interrupted')
                    raise

            local_prof.prof('start termination', uid=self._uid)

            self._logger.info('Terminating enqueue-thread')
            self._enqueue_thread_terminate.set()
            self._enqueue_thread.join()
            self._logger.info('Terminating dequeue-thread')
            self._dequeue_thread_terminate.set()
            self._dequeue_thread.join()

            local_prof.prof('termination done', uid=self._uid)

            local_prof.prof('terminating wfp process', uid=self._uid)

            local_prof.close()

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel wfprocessor process gracefully...')

            if self._enqueue_thread:

                if not self._enqueue_thread_terminate.is_set():
                    self._logger.info('Terminating enqueue-thread')
                    self._enqueue_thread_terminate.set()
                    self._enqueue_thread.join()

            if self._dequeue_thread:

                if not self._dequeue_thread_terminate.is_set():
                    self._logger.info('Terminating dequeue-thread')
                    self._dequeue_thread_terminate.set()
                    self._dequeue_thread.join()

            self._logger.info('WFprocessor process terminated')

            raise KeyboardInterrupt

        except Exception, ex:
            self._logger.error('Error in wfp process: %s. \n Closing enqueue, dequeue threads' % ex)

            if self._enqueue_thread:

                if not self._enqueue_thread_terminate.is_set():
                    self._logger.info('Terminating enqueue-thread')
                    self._enqueue_thread_terminate.set()
                    self._enqueue_thread.join()

            if self._dequeue_thread:

                if not self._dequeue_thread_terminate.is_set():
                    self._logger.info('Terminating dequeue-thread')
                    self._dequeue_thread_terminate.set()
                    self._dequeue_thread.join()

            self._logger.info('WFprocessor process terminated')

            print traceback.format_exc()
            raise Error(text=ex)

    def _enqueue(self, local_prof):
        """
        **Purpose**: This is the function that is run in the enqueue thread. This function extracts Tasks from the 
        copy of workflow that exists in the WFprocessor object and pushes them to the queues in the pending_q list.
        Since this thread works on the copy of the workflow, every state update to the Task, Stage and Pipeline is
        communicated back to the AppManager (master process) via the 'sync_with_master' function that has dedicated
        queues to communicate with the master.

        Details: Termination condition of this thread is set by the wfp process.
        """

        try:

            local_prof.prof('enqueue-thread started', uid=self._uid)
            self._logger.info('enqueue-thread started')

            if os.environ.get('DISABLE_RMQ_HEARTBEAT', None):
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port,
                                                                                        hearbeat=0
                                                                                        )
                                                              )
            else:
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port
                                                                                        )
                                                              )
            mq_channel = self._mq_connection.channel()

            while not self._enqueue_thread_terminate.is_set():

                for pipe in self._workflow:

                    with pipe._stage_lock:

                        if not pipe.completed:

                            # Test if the pipeline is already in the final state
                            if pipe.state in states.FINAL:
                                continue

                            elif pipe.state == states.INITIAL:

                                # Set state of pipeline to SCHEDULING if it is in INITIAL
                                transition(obj=pipe,
                                           obj_type='Pipeline',
                                           new_state=states.SCHEDULING,
                                           channel=mq_channel,
                                           queue='%s-enq-to-sync'%self._sid,
                                           profiler=local_prof,
                                           logger=self._logger)

                            if pipe.stages[pipe.current_stage - 1].state in [states.INITIAL, states.SCHEDULED]:


                                try:

                                    # Starting scheduling of tasks of current stage, so set state of stage to
                                    # SCHEDULING
                                    executable_stage = pipe.stages[pipe.current_stage - 1]

                                    if executable_stage.state == states.INITIAL:

                                        transition(obj=executable_stage,
                                                   obj_type='Stage',
                                                   new_state=states.SCHEDULING,
                                                   channel=mq_channel,
                                                   queue='%s-enq-to-sync'%self._sid,
                                                   profiler=local_prof,
                                                   logger=self._logger)

                                    executable_tasks = executable_stage.tasks

                                    for executable_task in executable_tasks:

                                        if self._resubmit_failed:

                                            if executable_task.state in [states.INITIAL, states.FAILED]:

                                                # Set state of Tasks in current Stage to SCHEDULING
                                                transition(obj=executable_task,
                                                           obj_type='Task',
                                                           new_state=states.SCHEDULING,
                                                           channel=mq_channel,
                                                           queue='%s-enq-to-sync'%self._sid,
                                                           profiler=local_prof,
                                                           logger=self._logger)

                                                task_as_dict = json.dumps(executable_task.to_dict())

                                                self._logger.debug('Publishing task %s to %s'
                                                                   % (executable_task.uid,
                                                                      self._pending_queue[0]))

                                                # Put the task on one of the pending_queues
                                                mq_channel.basic_publish(exchange='',
                                                                         routing_key=self._pending_queue[0],
                                                                         body=task_as_dict
                                                                         # properties=pika.BasicProperties(
                                                                         # make message persistent
                                                                         # delivery_mode = 2)
                                                                         )

                                                # Set state of Tasks in current Stage to SCHEDULED
                                                transition(obj=executable_task,
                                                           obj_type='Task',
                                                           new_state=states.SCHEDULED,
                                                           channel=mq_channel,
                                                           queue='%s-enq-to-sync'%self._sid,
                                                           profiler=local_prof,
                                                           logger=self._logger)

                                                tasks_submitted = True
                                                self._logger.debug('Task %s published to queue' % executable_task.uid)

                                        if executable_task.state == states.INITIAL:

                                            # Set state of Tasks in current Stage to SCHEDULING
                                            transition(obj=executable_task,
                                                       obj_type='Task',
                                                       new_state=states.SCHEDULING,
                                                       channel=mq_channel,
                                                       queue='%s-enq-to-sync'%self._sid,
                                                       profiler=local_prof,
                                                       logger=self._logger)

                                            task_as_dict = json.dumps(executable_task.to_dict())

                                            self._logger.debug('Publishing task %s to %s'
                                                               % (executable_task.uid,
                                                                  self._pending_queue[0]))

                                            # Put the task on one of the pending_queues
                                            mq_channel.basic_publish(exchange='',
                                                                     routing_key=self._pending_queue[0],
                                                                     body=task_as_dict
                                                                     # properties=pika.BasicProperties(
                                                                     # make message persistent
                                                                     # delivery_mode = 2)
                                                                     )

                                            # Set state of Tasks in current Stage to SCHEDULED
                                            transition(obj=executable_task,
                                                       obj_type='Task',
                                                       new_state=states.SCHEDULED,
                                                       channel=mq_channel,
                                                       queue='%s-enq-to-sync'%self._sid,
                                                       profiler=local_prof,
                                                       logger=self._logger)

                                            self._logger.debug('Task %s published to queue' % executable_task.uid)

                                    if executable_stage.state == states.SCHEDULING:
                                        # All tasks of current stage scheduled, so set state of stage to
                                        # SCHEDULED
                                        transition(obj=executable_stage,
                                                   obj_type='Stage',
                                                   new_state=states.SCHEDULED,
                                                   channel=mq_channel,
                                                   queue='%s-enq-to-sync'%self._sid,
                                                   profiler=local_prof,
                                                   logger=self._logger)
                                except Exception, ex:

                                    # If there is an error, explicitly switching the state of Stage to INITIAL
                                    self._logger.error('Error while updating stage ' +
                                                       'state, rolling back. Error: %s' % ex)

                                    executable_stage = pipe.stages[pipe.current_stage - 1]

                                    transition(obj=executable_stage,
                                               obj_type='Stage',
                                               new_state=states.INITIAL,
                                               channel=mq_channel,
                                               queue='%s-enq-to-sync'%self._sid,
                                               profiler=local_prof,
                                               logger=self._logger)

                                    raise

            self._logger.info('Enqueue thread terminated')
            self._mq_connection.close()

            local_prof.prof('terminating enqueue-thread', uid=self._uid)

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel enqueuer thread gracefully...')

            self._mq_connection.close()

            raise KeyboardInterrupt

        except Exception, ex:

            self._logger.error('Error in enqueue-thread: %s' % ex)
            print traceback.format_exc()

            raise Error(text=ex)

    def _dequeue(self, local_prof):
        """
        **Purpose**: This is the function that is run in the dequeue thread. This function extracts Tasks from the 
        completed queus and updates the copy of workflow that exists in the WFprocessor object.
        Since this thread works on the copy of the workflow, every state update to the Task, Stage and Pipeline is
        communicated back to the AppManager (master process) via the 'sync_with_master' function that has dedicated
        queues to communicate with the master.

        Details: Termination condition of this thread is set by the wfp process.
        """

        try:

            local_prof.prof('dequeue-thread started', uid=self._uid)
            self._logger.info('Dequeue thread started')

            if os.environ.get('DISABLE_RMQ_HEARTBEAT', None):
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port,
                                                                                        hearbeat=0
                                                                                        )
                                                              )
            else:
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port
                                                                                        )
                                                              )
            mq_channel = self._mq_connection.channel()

            while not self._dequeue_thread_terminate.is_set():

                try:

                    method_frame, header_frame, body = mq_channel.basic_get(queue=self._completed_queue[0])

                    if body:

                        # Get task from the message
                        completed_task = Task(duplicate=True)
                        completed_task.from_dict(json.loads(body))
                        self._logger.info('Got finished task %s from queue' % (completed_task.uid))

                        transition(obj=completed_task,
                                   obj_type='Task',
                                   new_state=states.DEQUEUEING,
                                   channel=mq_channel,
                                   queue='%s-deq-to-sync'%self._sid,
                                   profiler=local_prof,
                                   logger=self._logger)

                        # Traverse the entire workflow to find out the correct Task
                        for pipe in self._workflow:

                            with pipe._stage_lock:

                                if not pipe.completed:

                                    if completed_task.parent_pipeline == pipe.uid:

                                        self._logger.debug('Found parent pipeline: %s' % pipe.uid)

                                        for stage in pipe.stages:

                                            if completed_task.parent_stage == stage.uid:
                                                self._logger.debug('Found parent stage: %s' % (stage.uid))

                                                transition(obj=completed_task,
                                                           obj_type='Task',
                                                           new_state=states.DEQUEUED,
                                                           channel=mq_channel,
                                                           queue='%s-deq-to-sync'%self._sid,
                                                           profiler=local_prof,
                                                           logger=self._logger)

                                                if completed_task.exit_code:
                                                    completed_task.state = states.FAILED
                                                else:
                                                    completed_task.state = states.DONE

                                                for task in stage.tasks:

                                                    if task.uid == completed_task.uid:
                                                        task.state = str(completed_task.state)

                                                        if (task.state == states.FAILED) and (self._resubmit_failed):
                                                            task.state = states.INITIAL

                                                        transition(obj=task,
                                                                   obj_type='Task',
                                                                   new_state=task.state,
                                                                   channel=mq_channel,
                                                                   queue='%s-deq-to-sync'%self._sid,
                                                                   profiler=local_prof,
                                                                   logger=self._logger)

                                                        if stage._check_stage_complete():

                                                            transition(obj=stage,
                                                                       obj_type='Stage',
                                                                       new_state=states.DONE,
                                                                       channel=mq_channel,
                                                                       queue='%s-deq-to-sync'%self._sid,
                                                                       profiler=local_prof,
                                                                       logger=self._logger)

                                                            pipe._increment_stage()

                                                            if pipe.completed:

                                                                transition(obj=pipe,
                                                                           obj_type='Pipeline',
                                                                           new_state=states.DONE,
                                                                           channel=mq_channel,
                                                                           queue='%s-deq-to-sync'%self._sid,
                                                                           profiler=local_prof,
                                                                           logger=self._logger)

                                                        # Found the task and processed it -- no more iterations needed

                                                        break

                                                # Found the stage and processed it -- no more iterations neeeded
                                                break

                                        # Found the pipeline and processed it -- no more iterations neeeded
                                        break

                        mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                except Exception, ex:
                    self._logger.error('Unable to receive message from completed queue: %s' % ex)
                    raise

            self._logger.info('Terminated dequeue thread')
            self._mq_connection.close()

            local_prof.prof('terminating dequeue-thread', uid=self._uid)

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to exit gracefully...')

            self._mq_connection.close()

            raise KeyboardInterrupt

        except Exception, ex:
            self._logger.error('Error in dequeue-thread: %s' % ex)
            print traceback.format_exc()

            self._mq_connection.close()

            raise Error(text=ex)

    # ------------------------------------------------------------------------------------------------------------------
    # Public Methods
    # ------------------------------------------------------------------------------------------------------------------

    def start_processor(self):
        """
        **Purpose**: Method to start the wfp process. The wfp function
        is not to be accessed directly. The function is started in a separate
        process using this method.
        """

        if not self._wfp_process:

            try:

                self._prof.prof('creating wfp process', uid=self._uid)
                self._wfp_process = Process(target=self._wfp, name='wfprocessor')

                self._enqueue_thread = None
                self._dequeue_thread = None
                self._enqueue_thread_terminate = threading.Event()
                self._dequeue_thread_terminate = threading.Event()

                self._wfp_terminate = Event()
                self._logger.info('Starting WFprocessor process')
                self._prof.prof('starting wfp process', uid=self._uid)
                self._wfp_process.start()

                return True

            except Exception, ex:

                self._logger.error('WFprocessor not started')
                self.end_processor()
                raise

        else:
            self._logger.warn('Wfp process already running, attempted to restart!')

    def end_processor(self):
        """
        **Purpose**: Method to terminate the wfp process. This method is 
        blocking as it waits for the wfp process to terminate (aka join).
        """

        try:

            self._logger.debug('Attempting to end WFprocessor... event: %s' % self._wfp_terminate.is_set())

            if self.check_alive():
                self._wfp_terminate.set()
                self._wfp_process.join()
                self._logger.debug('WFprocessor process terminated')
            else:
                self._logger.debug('WFprocessor process already terminated')

            self._prof.prof('wfp process terminated', uid=self._uid)

            self._prof.close()

        except Exception, ex:
            self._logger.error('Could not terminate wfprocessor process')
            raise

    def workflow_incomplete(self):
        """
        **Purpose**: Method to check if the workflow execution is incomplete.
        """

        try:
            for pipe in self._workflow:
                with pipe._stage_lock:
                    if pipe.completed:
                        pass
                    else:
                        return True
            return False

        except Exception, ex:
            self._logger.error('Could not check if workflow is incomplete, error:%s' % ex)
            raise

    def check_alive(self):
        """
        **Purpose**: Method to check if the wfp process is alive
        """

        return self._wfp_process.is_alive()




__copyright__   = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__      = "Vivek Balasubramanian <vivek.balasubramanian@rutgers.edu>"
__license__     = "MIT"


# -----------------------------------------------------------------------------
# common states - Pipeline, Stage, Task
INITIAL         = 'DESCRIBED'
SCHEDULING      = 'SCHEDULING'

# common states - Stage, Task
SCHEDULED       = 'SCHEDULED'

# unique states - Tasks
SUBMITTING      = 'SUBMITTING'
SUBMITTED       = 'SUBMITTED'
COMPLETED       = 'EXECUTED'
DEQUEUEING      = 'DEQUEUEING'
DEQUEUED        = 'DEQUEUED'        # Dequeue thread will have this state
SYNCHRONIZING   = 'SYNCHRONIZING'
SYNCHRONIZED    = 'SYNCHRONIZED'    # Syncrhonizer thread will have this state

# common states - Pipeline, Stage, Task
DONE            = 'DONE'
FAILED          = 'FAILED'
CANCELED        = 'CANCELED'

# shortcut
FINAL = [DONE, FAILED, CANCELED]


## Assign numeric values to states
state_numbers = {
    
    INITIAL         : 1,
    SCHEDULING      : 2,
    SCHEDULED       : 3,
    SUBMITTING      : 4,
    SUBMITTED       : 5,
    COMPLETED       : 6,
    DEQUEUEING      : 7,
    DEQUEUED        : 8,
    SYNCHRONIZING   : 9,
    SYNCHRONIZED    : 10,
    DONE            : 11,
    FAILED          : 11,
    CANCELED        : 11
}

## Get back string values from numeric values for states
state_strings = {}
for k,v in state_numbers.iteritems():
    state_strings[v] = k


__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
from radical.entk.pipeline.pipeline import Pipeline
from radical.entk.stage.stage import Stage
from radical.entk.task.task import Task
from radical.entk.execman.resource_manager import ResourceManager
from radical.entk.execman.task_manager import TaskManager
from wfprocessor import WFprocessor
import sys
import time
import os
import Queue
import pika
import json
from threading import Thread, Event
import traceback
from radical.entk import states


class AppManager(object):

    """
    An application manager takes the responsibility of setting up the communication infrastructure, instantiates the
    ResourceManager, TaskManager, WFProcessor objects and all their threads and processes. This is the Master object
    running in the main process and is designed to recover from errors from all other objects, threads and processes.

    :Arguments:
        :hostname: host rabbitmq server is running
        :port: port at which rabbitmq can be accessed
        :reattempts: number of attempts to re-invoke any failed EnTK components
        :resubmit_failed: resubmit failed tasks (True/False)
        :autoterminate: terminate resource reservation upon execution of all tasks of first workflow (True/False)
    """

    def __init__(self,
                 hostname='localhost',
                 port=5672,
                 reattempts=3,
                 resubmit_failed=False,
                 autoterminate=True):

        # Create a session for each EnTK script execution
        self._sid = ru.generate_id('re.session', ru.ID_PRIVATE)

        # Create a folder for each session to hold EnTK profiles
        path = os.getcwd() + '/' + self._sid
        if not os.path.exists(path):
            os.makedirs(path)

        # Create an uid + logger + profiles for AppManager, under the sid
        # namespace
        self._uid = ru.generate_id('radical.entk.appmanager.%(item_counter)04d',
                                   ru.ID_CUSTOM,
                                   namespace=self._sid)
        self._logger = ru.get_logger(self._uid,
                                     path=path)
        self._prof = ru.Profiler(name=self._uid,
                                 path=path)

        self._prof.prof('create amgr obj', uid=self._uid)

        self._name = str()

        # RabbitMQ inits
        self._mq_hostname = hostname
        self._port = port

        # RabbitMQ Queues
        self._num_pending_qs = 1
        self._num_completed_qs = 1
        self._pending_queue = list()
        self._completed_queue = list()

        # Threads and procs counts
        self._num_push_threads = 1
        self._num_pull_threads = 1
        self._num_sync_threads = 1

        # Global parameters to have default values
        self._mqs_setup = False
        self._resource_manager = None
        self._task_manager = None
        self._workflow  = None
        self._resubmit_failed = resubmit_failed
        self._reattempts = reattempts
        self._cur_attempt = 1
        self._resource_autoterminate = autoterminate

        # Check if RP Profiler is set
        self._rp_profile = os.environ.get('RADICAL_PILOT_PROFILE', False)

        self._logger.info('Application Manager initialized')
        self._prof.prof('amgr obj created', uid=self._uid)

    # ------------------------------------------------------------------------------------------------------------------
    # Getter functions
    # ------------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        """
        Name for the application manager 

        :getter: Returns the name of the application manager
        :setter: Assigns the name of the application manager
        :type: String
        """

        return self._name

    @property
    def sid(self):
        """
        Get the session ID of the current EnTK execution

        :getter: Returns the session ID of the EnTK execution
        :type: String
        """

        return self._sid

    @property
    def resource_manager(self):
        """
        :getter: Returns the resource manager object being used
        :setter: Assigns a resource manager
        """

        return self._resource_manager

    # ------------------------------------------------------------------------------------------------------------------
    # Setter functions
    # ------------------------------------------------------------------------------------------------------------------

    @name.setter
    def name(self, value):

        if not isinstance(value, str):
            raise TypeError(expected_type=str, actual_type=type(value))

        else:
            self._name = value

    @resource_manager.setter
    def resource_manager(self, value):

        if not isinstance(value, ResourceManager):
            raise TypeError(expected_type=ResourceManager,
                            actual_type=type(value))
        else:
            self._resource_manager = value

            if self._resource_manager._validate_resource_desc(self._sid):
                self._resource_manager._populate()
            else:
                self._logger.error('Could not validate resource description')
                raise

    # ------------------------------------------------------------------------------------------------------------------
    # Public methods
    # ------------------------------------------------------------------------------------------------------------------

    def assign_workflow(self, workflow):
        """
        **Purpose**: Assign workflow to the application manager to be executed

        :arguments: set of Pipelines
        """

        self._prof.prof('assigning workflow', uid=self._uid)
        self._workflow = workflow
        self._logger.info('Workflow assigned to Application Manager')

 
    def run(self):
        """
        **Purpose**: Run the application manager. Once the workflow and resource manager have been assigned. Invoking this
        method will start the setting up the communication infrastructure, submitting a resource request and then
        submission of all the tasks.
        """

        try:

            # Set None objects local to each run
            self._wfp = None
            self._sync_thread = None
            self._end_sync = Event()
            self._resubmit_failed = False
            self._cur_attempt = 1

            if not self._workflow:
                self._logger.error('No workflow assigned currently, please check your script')
                raise MissingError(obj=self._uid, missing_attribute='workflow')

            if not self._resource_manager:
                self._logger.error(
                    'No resource manager assigned currently, please create and add a valid resource manager')
                raise MissingError(obj=self._uid, missing_attribute='resource_manager')

            else:

                self._prof.prof('amgr run started', uid=self._uid)

                # Setup rabbitmq stuff
                if not self._mqs_setup:

                    self._logger.info('Setting up RabbitMQ system')
                    setup = self._setup_mqs()

                    if not setup:
                        self._logger.error('RabbitMQ system not available')
                        raise Error(text="RabbitMQ setup failed")

                    self._mqs_setup = True


                # Create WFProcessor object
                self._prof.prof('creating wfp obj', uid=self._uid)
                self._wfp = WFprocessor(sid=self._sid,
                                        workflow=self._workflow,
                                        pending_queue=self._pending_queue,
                                        completed_queue=self._completed_queue,
                                        mq_hostname=self._mq_hostname,
                                        port=self._port,
                                        resubmit_failed=self._resubmit_failed)
                self._wfp._validate_workflow()
                self._workflow = self._wfp.workflow


                # Submit resource request if not resource allocation done till now or
                # resubmit a new one if the old one has completed
                if self._resource_manager:
                    res_alloc_state = self._resource_manager.get_resource_allocation_state()
                    if (not res_alloc_state) or (res_alloc_state in self._resource_manager.completed_states()):

                        self._logger.info('Starting resource request submission')
                        self._prof.prof('init rreq submission', uid=self._uid)
                        self._resource_manager._submit_resource_request()

                else:

                    self._logger.error(
                        'Cannot run without resource manager, please create and assign a resource manager')
                    raise Error(text='Missing resource manager')

                # Start synchronizer thread
                if not self._sync_thread:
                    self._logger.info('Starting synchronizer thread')
                    self._sync_thread = Thread(target=self._synchronizer, name='synchronizer-thread')
                    self._prof.prof('starting synchronizer thread', uid=self._uid)
                    self._sync_thread.start()

                
                # Start WFprocessor
                self._logger.info('Starting WFProcessor process from AppManager')
                self._wfp.start_processor()

                # Create tmgr object only if it does not already exist
                if not self._task_manager:
                    self._prof.prof('creating tmgr obj', uid=self._uid)
                    self._task_manager = TaskManager(sid=self._sid,
                                                     pending_queue=self._pending_queue,
                                                     completed_queue=self._completed_queue,
                                                     mq_hostname=self._mq_hostname,
                                                     rmgr=self._resource_manager,
                                                     port=self._port
                                                     )
                    self._logger.info('Starting task manager process from AppManager')
                    self._task_manager.start_manager()
                    self._task_manager.start_heartbeat()

                active_pipe_count = len(self._workflow)
                finished_pipe_uids = []

                # We wait till all pipelines of the workflow are marked
                # complete
                while ( (active_pipe_count > 0) and
                        (self._wfp.workflow_incomplete()) and
                        (self._resource_manager.get_resource_allocation_state() not
                        in self._resource_manager.completed_states())):

                    if active_pipe_count > 0:

                        for pipe in self._workflow:

                            with pipe._stage_lock:

                                if (pipe.completed) and (pipe.uid not in finished_pipe_uids):

                                    self._logger.info('Pipe %s completed' % pipe.uid)
                                    finished_pipe_uids.append(pipe.uid)
                                    active_pipe_count -= 1
                                    self._logger.info('Active pipes: %s' % active_pipe_count)

                    if (not self._sync_thread.is_alive()) and (self._cur_attempt <= self._reattempts):

                        self._sync_thread = Thread(target=self._synchronizer,
                                                   name='synchronizer-thread')
                        self._logger.info('Restarting synchronizer thread')
                        self._prof.prof('restarting synchronizer', uid=self._uid)
                        self._sync_thread.start()

                        self._cur_attempt += 1

                    if (not self._wfp.check_alive()) and (self._cur_attempt <= self._reattempts):

                        """
                        If WFP dies, both child threads are also cleaned out.
                        We simply recreate the wfp object with a copy of the workflow
                        in the appmanager and start the processor.
                        """

                        self._prof.prof('recreating wfp obj', uid=self._uid)
                        self._wfp = WFProcessor(
                            sid=self._sid,
                            workflow=self._workflow,
                            pending_queue=self._pending_queue,
                            completed_queue=self._completed_queue,
                            mq_hostname=self._mq_hostname,
                            port=self._port,
                            resubmit_failed=self._resubmit_failed)

                        self._logger.info('Restarting WFProcessor process from AppManager')
                        self._wfp.start_processor()

                        self._cur_attempt += 1

                    if (not self._task_manager.check_alive()) and (self._cur_attempt <= self._reattempts):

                        """
                        If the tmgr process dies, we simply start a new process
                        using the start_manager method. We do not need to create
                        a new instance of the TaskManager object itself. We stop
                        and start a new isntance of the heartbeat thread as well.
                        """
                        self._prof.prof('restarting tmgr process', uid=self._uid)

                        self._logger.info('Terminating heartbeat thread from AppManager')
                        self._task_manager.end_heartbeat()
                        self._logger.info('Restarting task manager process from AppManager')
                        self._task_manager.start_manager()
                        self._logger.info('Restarting heartbeat thread from AppManager')
                        self._task_manager.start_heartbeat()

                        self._cur_attempt += 1

                    if (not self._task_manager.check_heartbeat()) and (self._cur_attempt <= self._reattempts):

                        """
                        If the heartbeat thread dies, we simply start a new thread
                        using the start_heartbeat method. We do not need to create
                        a new instance of the TaskManager object itself. We stop
                        and start a new isntance of the tmgr process as well
                        """

                        self._prof.prof('restarting heartbeat thread', uid=self._uid)

                        self._logger.info('Terminating tmgr process from AppManager')
                        self._task_manager.end_manager()
                        self._logger.info('Restarting task manager process from AppManager')
                        self._task_manager.start_manager()
                        self._logger.info('Restarting heartbeat thread from AppManager')
                        self._task_manager.start_heartbeat()

                        self._cur_attempt += 1

                self._prof.prof('start termination', uid=self._uid)

                # Terminate threads in following order: wfp, helper, synchronizer
                self._logger.info('Terminating WFprocessor')
                self._wfp.end_processor()

                self._logger.info('Terminating synchronizer thread')
                self._end_sync.set()
                self._sync_thread.join()
                self._logger.info('Synchronizer thread terminated')

                if self._resource_autoterminate:

                    self._logger.info('Terminating task manager process')
                    self._task_manager.end_manager()
                    self._task_manager.end_heartbeat()

                    self._resource_manager._cancel_resource_request(self._rp_profile)

                self._prof.prof('termination done', uid=self._uid)

        except KeyboardInterrupt:

            self._prof.prof('start termination', uid=self._uid)

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel enqueuer thread gracefully...')

            # Terminate threads in following order: wfp, helper, synchronizer
            if self._wfp:
                self._logger.info('Terminating WFprocessor')
                self._wfp.end_processor()

            if self._task_manager:
                self._logger.info('Terminating task manager process')
                self._task_manager.end_manager()
                self._task_manager.end_heartbeat()

            if self._sync_thread:
                self._logger.info('Terminating synchronizer thread')
                self._end_sync.set()
                self._sync_thread.join()
                self._logger.info('Synchronizer thread terminated')

            if self._resource_manager:
                self._resource_manager._cancel_resource_request(self._rp_profile)

            self._prof.prof('termination done', uid=self._uid)

            raise KeyboardInterrupt

        except Exception, ex:

            self._prof.prof('start termination', uid=self._uid)

            self._logger.error('Error in AppManager')

            print traceback.format_exc()

            # Terminate threads in following order: wfp, helper, synchronizer
            if self._wfp:
                self._logger.info('Terminating WFprocessor')
                self._wfp.end_processor()

            if self._task_manager:
                self._logger.info('Terminating task manager process')
                self._task_manager.end_manager()
                self._task_manager.end_heartbeat()

            if self._sync_thread:
                self._logger.info('Terminating synchronizer thread')
                self._end_sync.set()
                self._sync_thread.join()
                self._logger.info('Synchronizer thread terminated')

            if self._resource_manager:
                self._resource_manager._cancel_resource_request(self._rp_profile)

            self._prof.prof('termination done', uid=self._uid)

            raise Error(text=ex)

    def resource_terminate(self):

        if self._task_manager:
            self._logger.info('Terminating task manager process')
            self._task_manager.end_manager()
            self._task_manager.end_heartbeat()

        if self._resource_manager:
            self._resource_manager._cancel_resource_request(self._rp_profile)

    # ------------------------------------------------------------------------------------------------------------------
    # Private methods
    # ------------------------------------------------------------------------------------------------------------------  

    def _setup_mqs(self):
        """
        **Purpose**: Setup RabbitMQ system on the client side. We instantiate queue(s) 'pendingq-*' for communication 
        between the enqueuer thread and the task manager process. We instantiate queue(s) 'completedq-*' for
        communication between the task manager and dequeuer thread. We instantiate queue 'sync-to-master' for 
        communication from enqueuer/dequeuer/task_manager to the synchronizer thread. We instantiate queue
        'sync-ack' for communication from synchronizer thread to enqueuer/dequeuer/task_manager.

        Details: All queues are durable: Even if the RabbitMQ server goes down, the queues are saved to disk and can
        be retrieved. This also means that after an erroneous run the queues might still have unacknowledged messages
        and will contain messages from that run. Hence, in every new run, we first delete the queue and create a new 
        one.
        """

        try:

            self._prof.prof('init mqs setup', uid=self._uid)

            self._logger.debug('Setting up mq connection and channel')

            if os.environ.get('DISABLE_RMQ_HEARTBEAT', None):
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port,
                                                                                        hearbeat=0
                                                                                        )
                                                              )
            else:
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port
                                                                                        )
                                                              )

            self._mq_channel = self._mq_connection.channel()

            self._logger.debug('Connection and channel setup successful')

            self._logger.debug('Setting up all exchanges and queues')

            for i in range(1, self._num_pending_qs + 1):
                queue_name = '%s-pendingq-%s' % (self._sid, i)
                self._pending_queue.append(queue_name)
                self._mq_channel.queue_delete(queue=queue_name)
                # Durable Qs will not be lost if rabbitmq server crashes
                self._mq_channel.queue_declare(queue=queue_name, durable=True)

            for i in range(1, self._num_completed_qs + 1):
                queue_name = '%s-completedq-%s' % (self._sid, i)
                self._completed_queue.append(queue_name)
                self._mq_channel.queue_delete(queue=queue_name)
                # Durable Qs will not be lost if rabbitmq server crashes
                self._mq_channel.queue_declare(queue=queue_name, durable=True)

            # self._mq_channel.queue_delete(queue='sync-to-master')
            # self._mq_channel.queue_declare(queue='sync-to-master')

            # Queues to send messages from the threads/procs to master
            self._mq_channel.queue_delete(queue='%s-tmgr-to-sync' % self._sid)
            self._mq_channel.queue_declare(queue='%s-tmgr-to-sync' % self._sid)
            self._mq_channel.queue_delete(queue='%s-cb-to-sync' % self._sid)
            self._mq_channel.queue_declare(queue='%s-cb-to-sync' % self._sid)
            self._mq_channel.queue_delete(queue='%s-enq-to-sync' % self._sid)
            self._mq_channel.queue_declare(queue='%s-enq-to-sync' % self._sid)
            self._mq_channel.queue_delete(queue='%s-deq-to-sync' % self._sid)
            self._mq_channel.queue_declare(queue='%s-deq-to-sync' % self._sid)

            # Queues to send messages from master to threads/procs
            self._mq_channel.queue_delete(queue='%s-sync-to-tmgr' % self._sid)
            self._mq_channel.queue_declare(queue='%s-sync-to-tmgr' % self._sid)
            self._mq_channel.queue_delete(queue='%s-sync-to-cb' % self._sid)
            self._mq_channel.queue_declare(queue='%s-sync-to-cb' % self._sid)
            self._mq_channel.queue_delete(queue='%s-sync-to-enq' % self._sid)
            self._mq_channel.queue_declare(queue='%s-sync-to-enq' % self._sid)
            self._mq_channel.queue_delete(queue='%s-sync-to-deq' % self._sid)
            self._mq_channel.queue_declare(queue='%s-sync-to-deq' % self._sid)
            # Durable Qs will not be lost if rabbitmq server crashes

            self._logger.debug('All exchanges and queues are setup')
            self._prof.prof('mqs setup done', uid=self._uid)

            return True

        except Exception, ex:

            self._logger.error('Error setting RabbitMQ system: %s' % ex)
            raise

    def _synchronizer(self):
        """
        **Purpose**: Thread in the master process to keep the workflow data 
        structure in appmanager up to date. We receive pipelines, stages and 
        tasks objects directly. The respective object is updated in this master 
        process. 

        Details: Important to note that acknowledgements of the type
        channel.basic_ack() is an acknowledgement to the server that the msg
        was received. This is not to be confused with the Ack sent to the 
        enqueuer/dequeuer/task_manager through the sync-ack queue. 
        """

        try:

            self._prof.prof('synchronizer started', uid=self._uid)

            self._logger.info('synchronizer thread started')

            def task_update(msg, reply_to, corr_id, mq_channel):

                completed_task = Task(duplicate=True)
                completed_task.from_dict(msg['object'])
                self._logger.info('Received %s with state %s' % (completed_task.uid, completed_task.state))

                # Traverse the entire workflow to find the correct task
                for pipe in self._workflow:

                    with pipe._stage_lock:

                        if not pipe.completed:
                            if completed_task.parent_pipeline == pipe.uid:

                                for stage in pipe.stages:

                                    if completed_task.parent_stage == stage.uid:

                                        for task in stage.tasks:

                                            if (completed_task.uid == task.uid)and(completed_task.state != task.state):

                                                task.state = str(completed_task.state)
                                                self._logger.debug('Found task %s with state %s' %
                                                                   (task.uid, task.state))

                                                if completed_task.path:
                                                    task.path = str(completed_task.path)

                                                print 'Syncing task %s with state %s' % (task.uid, task.state)

                                                mq_channel.basic_publish(exchange='',
                                                                         routing_key=reply_to,
                                                                         properties=pika.BasicProperties(
                                                                             correlation_id=corr_id),
                                                                         body='%s-ack' % task.uid)

                                                self._prof.prof('publishing sync ack for obj with state %s' %
                                                                msg['object']['state'],
                                                                uid=msg['object']['uid']
                                                                )

                                                mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                                                print 'Synced task %s with state %s' % (task.uid, task.state)

            def stage_update(msg, reply_to, corr_id, mq_channel):

                completed_stage = Stage(duplicate=True)
                completed_stage.from_dict(msg['object'])
                self._logger.info('Received %s with state %s' % (completed_stage.uid, completed_stage.state))

                # Traverse the entire workflow to find the correct stage
                for pipe in self._workflow:

                    with pipe._stage_lock:

                        if not pipe.completed:
                            if completed_stage.parent_pipeline == pipe.uid:
                                self._logger.info('Found parent pipeline: %s' % pipe.uid)

                                for stage in pipe.stages:

                                    if (completed_stage.uid == stage.uid)and(completed_stage.state != stage.state):

                                        self._logger.debug('Found stage %s' % stage.uid)

                                        stage.state = str(completed_stage.state)

                                        mq_channel.basic_publish(exchange='',
                                                                 routing_key=reply_to,
                                                                 properties=pika.BasicProperties(
                                                                     correlation_id=corr_id),
                                                                 body='%s-ack' % stage.uid)

                                        self._prof.prof('publishing sync ack for obj with state %s' %
                                                        msg['object']['state'],
                                                        uid=msg['object']['uid']
                                                        )

                                        mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

            def pipeline_update(msg, reply_to, corr_id, mq_channel):

                completed_pipeline = Pipeline()
                completed_pipeline.from_dict(msg['object'])

                self._logger.info('Received %s with state %s' % (completed_pipeline.uid, completed_pipeline.state))

                # Traverse the entire workflow to find the correct pipeline
                for pipe in self._workflow:

                    with pipe._stage_lock:

                        if not pipe.completed:

                            if (completed_pipeline.uid == pipe.uid)and(completed_pipeline.state != pipe.state):

                                pipe.state = str(completed_pipeline.state)

                                self._logger.info('Found pipeline %s, state %s, completed %s' % (pipe.uid,
                                                                                                 pipe.state,
                                                                                                 pipe.completed)
                                                  )

                                # Reply with ack msg to the sender
                                mq_channel.basic_publish(exchange='',
                                                         routing_key=reply_to,
                                                         properties=pika.BasicProperties(
                                                             correlation_id=corr_id),
                                                         body='%s-ack' % pipe.uid)

                                self._prof.prof('publishing sync ack for obj with state %s' %
                                                msg['object']['state'],
                                                uid=msg['object']['uid']
                                                )

                                mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                                # Keep the assignment of the completed flag after sending the acknowledgment
                                # back. Otherwise the MainThread takes lock over the pipeline because of logging
                                # and profiling
                                if completed_pipeline.completed:
                                    pipe._completed_flag.set()

            # Disable heartbeat for long running jobs since that might load the TCP channel
            # https://github.com/pika/pika/issues/753
            if os.environ.get('DISABLE_RMQ_HEARTBEAT', None):
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port,
                                                                                        hearbeat=0
                                                                                        )
                                                              )
            else:
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port
                                                                                        )
                                                              )
            mq_channel = self._mq_connection.channel()

            while not self._end_sync.is_set():

                #-------------------------------------------------------------------------------------------------------
                # Messages between tmgr Main thread and synchronizer -- only Task objects

                method_frame, props, body = mq_channel.basic_get(queue='%s-tmgr-to-sync'%self._sid)

                """
                The message received is a JSON object with the following structure:

                msg = {
                        'type': 'Pipeline'/'Stage'/'Task',
                        'object': json/dict
                        }
                """

                if body:

                    msg = json.loads(body)

                    self._prof.prof('received obj with state %s for sync' %
                                    msg['object']['state'], uid=msg['object']['uid'])

                    if msg['type'] == 'Task':
                        task_update(msg, '%s-sync-to-tmgr'%self._sid, props.correlation_id, mq_channel)

                #-------------------------------------------------------------------------------------------------------

                #-------------------------------------------------------------------------------------------------------
                # Messages between callback thread and synchronizer -- only Task objects

                method_frame, props, body = mq_channel.basic_get(queue='%s-cb-to-sync'%self._sid)

                """
                The message received is a JSON object with the following structure:

                msg = {
                        'type': 'Pipeline'/'Stage'/'Task',
                        'object': json/dict
                        }
                """

                if body:

                    msg = json.loads(body)

                    self._prof.prof('received obj with state %s for sync' %
                                    msg['object']['state'], uid=msg['object']['uid'])

                    if msg['type'] == 'Task':
                        task_update(msg, '%s-sync-to-cb'%self._sid, props.correlation_id, mq_channel)

                #-------------------------------------------------------------------------------------------------------

                #-------------------------------------------------------------------------------------------------------
                # Messages between enqueue thread and synchronizer -- Task, Stage or Pipeline
                method_frame, props, body = mq_channel.basic_get(queue='%s-enq-to-sync'%self._sid)

                if body:

                    msg = json.loads(body)

                    self._prof.prof('received obj with state %s for sync' %
                                    msg['object']['state'], uid=msg['object']['uid'])

                    if msg['type'] == 'Task':
                        task_update(msg, '%s-sync-to-enq'%self._sid, props.correlation_id, mq_channel)

                    elif msg['type'] == 'Stage':
                        stage_update(msg, '%s-sync-to-enq'%self._sid, props.correlation_id, mq_channel)

                    elif msg['type'] == 'Pipeline':
                        pipeline_update(msg, '%s-sync-to-enq'%self._sid, props.correlation_id, mq_channel)
                #-------------------------------------------------------------------------------------------------------

                #-------------------------------------------------------------------------------------------------------
                # Messages between dequeue thread and synchronizer -- Task, Stage or Pipeline
                method_frame, props, body = mq_channel.basic_get(queue='%s-deq-to-sync'%self._sid)

                if body:

                    msg = json.loads(body)

                    self._prof.prof('received obj with state %s for sync' %
                                    msg['object']['state'], uid=msg['object']['uid'])

                    if msg['type'] == 'Task':
                        task_update(msg, '%s-sync-to-deq'%self._sid, props.correlation_id, mq_channel)

                    elif msg['type'] == 'Stage':
                        stage_update(msg, '%s-sync-to-deq'%self._sid, props.correlation_id, mq_channel)

                    elif msg['type'] == 'Pipeline':
                        pipeline_update(msg, '%s-sync-to-deq'%self._sid, props.correlation_id, mq_channel)
                #-------------------------------------------------------------------------------------------------------

            self._prof.prof('terminating synchronizer', uid=self._uid)

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to terminate synchronizer thread gracefully...')

            self._end_sync.set()
            raise KeyboardInterrupt

        except Exception, ex:

            self._logger.error('Unknown error in synchronizer: %s. \n Terminating thread' % ex)
            print traceback.format_exc()
            self._end_sync.set()
            raise Error(text=ex)

    # ------------------------------------------------------------------------------------------------------------------



from radical.entk import Pipeline, Stage, Task, AppManager, ResourceManager
import os

# ------------------------------------------------------------------------------
# Set default verbosity

if os.environ.get('RADICAL_ENTK_VERBOSE') == None:
    os.environ['RADICAL_ENTK_VERBOSE'] = 'INFO'

if __name__ == '__main__':

    # Create a Pipeline object
    p = Pipeline()

    # Create a Stage object
    s = Stage()

    for cnt in range(10):

        # Create a Task object
        t = Task()
        t.name = 'my-task'        # Assign a name to the task (optional)
        t.executable = ['/bin/echo']   # Assign executable to the task   
        t.arguments = ['I am task %s'%cnt]  # Assign arguments for the task executable

        # Add the Task to the Stage
        s.add_tasks(t)

    # Add Stage to the Pipeline
    p.add_stages(s)


    # Create a dictionary describe four mandatory keys:
    # resource, walltime, cores and project
    # resource is 'local.localhost' to execute locally
    res_dict = {

            'resource': 'local.localhost',
            'walltime': 10,
            'cores': 1,
            'project': '',
    }

    # Create Resource Manager object with the above resource description
    rman = ResourceManager(res_dict)

    # Create Application Manager
    appman = AppManager()

    # Assign resource manager to the Application Manager
    appman.resource_manager = rman

    # Assign the workflow as a set of Pipelines to the Application Manager
    appman.assign_workflow(set([p]))

    # Run the Application Manager
    appman.run()



__copyright__ = "Copyright 2017-2018, http://radical.rutgers.edu"
__author__ = "Vivek Balasubramanian <vivek.balasubramaniana@rutgers.edu>"
__license__ = "MIT"

import radical.utils as ru
from radical.entk.exceptions import *
import threading
from multiprocessing import Process, Event
import Queue
from radical.entk import states, Task
from radical.entk.execman.resource_manager import ResourceManager
from radical.entk.utils.init_transition import transition
import time
import json
import pika
import traceback
import os
import radical.pilot as rp
from task_processor import create_cud_from_task, create_task_from_cu
import uuid


class TaskManager(object):

    """
    A Task Manager takes the responsibility of dispatching tasks it receives from a queue for execution on to 
    the available resources using a runtime system. In this case, the runtime system being used RADICAL Pilot. Once 
    the tasks have completed execution, they are pushed on to another queue for other components of EnTK to access.


    :arguments:
        :pending_queue: List of queue(s) with tasks ready to be executed. Currently, only one queue.
        :completed_queue: List of queue(s) with tasks that have finished execution. Currently, only one queue.
        :rmgr: ResourceManager object to be used to access the Pilot where the tasks can be submitted
        :mq_hostname: Name of the host where RabbitMQ is running
        :port: port at which rabbitMQ can be accessed

    Currently, EnTK is configured to work with one pending queue and one completed queue. In the future, the number of 
    queues can be varied for different throughput requirements at the cost of additional Memory and CPU consumption.
    """

    def __init__(self, sid, pending_queue, completed_queue, rmgr, mq_hostname, port):

        if isinstance(sid, str):
            self._sid = sid
        else:
            raise TypeError(expected_type=str, actual_type=type(sid))

        self._uid = ru.generate_id('radical.entk.task_manager.%(item_counter)04d', ru.ID_CUSTOM, namespace=self._sid)
        self._path = os.getcwd() + '/' + self._sid

        self._logger = ru.get_logger(self._uid, path=self._path)
        self._prof = ru.Profiler(name=self._uid + '-obj', path=self._path)

        self._prof.prof('create tmgr obj', uid=self._uid)

        if isinstance(pending_queue, list):
            self._pending_queue = pending_queue
        else:
            raise TypeError(expected_type=str, actual_type=type(pending_queue))

        if isinstance(completed_queue, list):
            self._completed_queue = completed_queue
        else:
            raise TypeError(expected_type=str, actual_type=type(completed_queue))

        if isinstance(mq_hostname, str):
            self._mq_hostname = mq_hostname
        else:
            raise TypeError(expected_type=str, actual_type=type(mq_hostname))

        if isinstance(port, int):
            self._port = port
        else:
            raise TypeError(expected_type=int, actual_type=type(port))

        if isinstance(rmgr, ResourceManager):
            self._rmgr = rmgr
        else:
            raise TypeError(expected_type=ResourceManager, actual_type=type(rmgr))

        self._tmgr_process = None
        self._tmgr_terminate = None
        self._hb_thread = None
        self._hb_alive = None
        self._umgr = None

        self._logger.info('Created task manager object: %s' % self._uid)

        self._prof.prof('tmgr obj created', uid=self._uid)

    # ------------------------------------------------------------------------------------------------------------------
    # Private Methods
    # ------------------------------------------------------------------------------------------------------------------

    def _heartbeat(self):
        """
        **Purpose**: Method to be executed in the heartbeat thread. This method sends a 'request' to the
        heartbeat-req queue. It expects a 'response' message from the 'heartbeart-res' queue within 10 seconds. This
        message should contain the same correlation id. If no message if received in 10 seconds, the tmgr is assumed
        dead. The end_manager() is called to cleanly terminate tmgr process and the heartbeat thread is also 
        terminated.

        **Details**: The AppManager can re-invoke both if the execution is still not complete.
        """

        try:

            self._prof.prof('heartbeat thread started', uid=self._uid)

            if os.environ.get('DISABLE_RMQ_HEARTBEAT', None):
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port,
                                                                                        hearbeat=0
                                                                                        )
                                                              )
            else:
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=self._mq_hostname,
                                                                                        port=self._port
                                                                                        )
                                                              )

            channel = self._mq_connection.channel()
            channel.queue_delete(queue='%s-heartbeat-req'%self._sid)
            channel.queue_declare(queue='%s-heartbeat-req'%self._sid)
            response = True

            while (response and (not self._hb_alive.is_set())):
                response = False
                corr_id = str(uuid.uuid4())

                # Heartbeat request signal sent to task manager via rpc-queue
                channel.basic_publish(exchange='',
                                      routing_key='%s-heartbeat-req'%self._sid,
                                      properties=pika.BasicProperties(
                                          reply_to='%s-heartbeat-res'%self._sid,
                                          correlation_id=corr_id),
                                      body='request')

                self._logger.info('Sent heartbeat request')

                # Ten second interval for heartbeat request to be responded to
                time.sleep(10)

                method_frame, props, body = channel.basic_get(queue='%s-heartbeat-res'%self._sid)

                if body:
                    if corr_id == props.correlation_id:
                        self._logger.info('Received heartbeat response')
                        response = True

                        channel.basic_ack(delivery_tag=method_frame.delivery_tag)

        except KeyboardInterrupt:
            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel tmgr process gracefully...')
            raise KeyboardInterrupt

        except Exception as ex:
            self._logger.error('Heartbeat failed with error: %s' % ex)
            raise

        finally:

            if self._hb_thread.is_alive():
                self._hb_alive.set()

            self._prof.prof('terminating hearbeat thread', uid=self._uid)

    def _tmgr(self, uid, umgr, rmgr, logger, mq_hostname, port, pending_queue, completed_queue):
        """
        **Purpose**: Method to be run by the tmgr process. This method receives a Task from the pending_queue
        and submits it to the RTS. Currently, it also converts Tasks into CUDs and CUs into (partially described) Tasks.
        This conversion is necessary since the current RTS is RADICAL Pilot. Once Tasks are recovered from a CU, they
        are then pushed to the completed_queue. At all state transititons, they are synced (blocking) with the AppManager
        in the master process.

        In addition the tmgr also receives heartbeat 'request' msgs from the heartbeat-req queue. It responds with a
        'response' message to the 'heartbeart-res' queue.

        **Details**: The AppManager can re-invoke the tmgr process with this function if the execution of the workflow is 
        still incomplete. There is also population of a dictionary, placeholder_dict, which stores the path of each of
        the tasks on the remote machine. 
        """

        try:

            local_prof = ru.Profiler(name=self._uid + '-proc', path=self._path)

            local_prof.prof('tmgr process started', uid=self._uid)
            logger.info('Task Manager process started')

            placeholder_dict = dict()

            def load_placeholder(task):

                parent_pipeline = str(task.parent_pipeline)
                parent_stage = str(task.parent_stage)

                if parent_pipeline not in placeholder_dict:
                    placeholder_dict[parent_pipeline] = dict()

                if parent_stage not in placeholder_dict[parent_pipeline]:
                    placeholder_dict[parent_pipeline][parent_stage] = dict()

                placeholder_dict[parent_pipeline][parent_stage][str(task.uid)] = str(task.path)

            def unit_state_cb(unit, state):

                try:

                    logger.debug('Unit %s in state %s' % (unit.uid, unit.state))

                    # Thread should run till terminate condtion is encountered
                    if os.environ.get('DISABLE_RMQ_HEARTBEAT', None):
                        self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=mq_hostname,
                                                                                                port=port,
                                                                                                hearbeat=0
                                                                                                )
                                                                      )
                    else:
                        self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=mq_hostname,
                                                                                                port=port
                                                                                                )
                                                                      )
                    mq_channel = self._mq_connection.channel()

                    if unit.state in [rp.DONE, rp.FAILED]:

                        try:

                            task = None
                            task = create_task_from_cu(unit, local_prof)

                            transition(obj=task,
                                       obj_type='Task',
                                       new_state=states.COMPLETED,
                                       channel=mq_channel,
                                       queue='%s-cb-to-sync'%self._sid,
                                       profiler=local_prof,
                                       logger=logger)

                            load_placeholder(task)

                        except Exception, ex:

                            # Rollback and pass exception
                            if task:
                                self._logger.error('Task %s creation for completed cu %s failed, error: %s' %
                                                   (task.uid, unit.uid, ex))
                                task.state = states.SCHEDULED
                                transition(obj=task,
                                           obj_type='Task',
                                           new_state=states.SCHEDULED,
                                           channel=mq_channel,
                                           queue='%s-cb-to-sync'%self._sid,
                                           profiler=local_prof,
                                           logger=logger)
                            else:
                                self._logger.error(
                                    'Task creation from completed cu %s failed, error: %s' % (unit.uid, ex))

                            raise

                        task_as_dict = json.dumps(task.to_dict())

                        mq_channel.basic_publish(exchange='',
                                                 routing_key='%s-completedq-1'%self._sid,
                                                 body=task_as_dict
                                                 # properties=pika.BasicProperties(
                                                 # make message persistent
                                                 #    delivery_mode = 2,
                                                 #)
                                                 )

                        logger.info('Pushed task %s with state %s to completed queue %s' % (
                            task.uid,
                            task.state,
                            completed_queue[0])
                        )

                    self._mq_connection.close()

                except KeyboardInterrupt:
                    self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                                       'trying to exit callback thread gracefully...')

                    raise KeyboardInterrupt

                except Exception, ex:

                    self._logger.error('Error in RP callback thread: %s' % ex)
                    print traceback.format_exc()
                    # raise      # Not necessary to raise the callback thread, we don't want RP to react

            if not umgr:
                umgr = rp.UnitManager(session=rmgr._session)
                umgr.add_pilots(rmgr.pilot)
                umgr.register_callback(unit_state_cb)

            # Thread should run till terminate condtion is encountered
            if os.environ.get('DISABLE_RMQ_HEARTBEAT', None):
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=mq_hostname,
                                                                                        port=port,
                                                                                        hearbeat=0
                                                                                        )
                                                              )
            else:
                self._mq_connection = pika.BlockingConnection(pika.ConnectionParameters(host=mq_hostname,
                                                                                        port=port
                                                                                        )
                                                              )
            mq_channel = self._mq_connection.channel()

            # To respond to heartbeat - get request from rpc_queue
            mq_channel.queue_delete(queue='%s-heartbeat-res'%self._sid)
            mq_channel.queue_declare(queue='%s-heartbeat-res'%self._sid)

            '''
            # Function to be invoked upon request message
            def on_request(ch, method, props, body):

                self._logger.info('Received heartbeat request')

                ch.basic_publish(   exchange='',
                                    routing_key=props.reply_to,
                                    properties=pika.BasicProperties(correlation_id = props.correlation_id),
                                    body='response')

                self._logger.info('Sent heartbeat response')

                ch.basic_ack(delivery_tag = method.delivery_tag)
            '''

            local_prof.prof('tmgr infrastructure setup done', uid=uid)

            while not self._tmgr_terminate.is_set():

                try:

                    method_frame, header_frame, body = mq_channel.basic_get(queue=pending_queue[0])

                    if body:

                        try:

                            task = None
                            task = Task(duplicate=True)
                            task.from_dict(json.loads(body))

                            transition(obj=task,
                                       obj_type='Task',
                                       new_state=states.SUBMITTING,
                                       channel=mq_channel,
                                       queue='%s-tmgr-to-sync'%self._sid,
                                       profiler=local_prof,
                                       logger=self._logger)

                        except Exception, ex:

                            # Rollback and pass exception
                            if task:
                                self._logger.error(
                                    'Task %s preparation for submission failed, error: %s' % (task.uid, ex))
                                task.state = states.SCHEDULED
                                transition(obj=task,
                                           obj_type='Task',
                                           new_state=states.SCHEDULED,
                                           channel=mq_channel,
                                           queue='%s-tmgr-to-sync'%self._sid,
                                           profiler=local_prof,
                                           logger=self._logger)
                            else:
                                self._logger.error('Task preparation for submission failed, error: %s' % ex)

                            raise

                        try:

                            umgr.submit_units(create_cud_from_task(task, placeholder_dict, local_prof))
                            self._logger.info('Task %s, %s; submitted to RTS' % (task.uid, task.state))

                            transition(obj=task,
                                       obj_type='Task',
                                       new_state=states.SUBMITTED,
                                       channel=mq_channel,
                                       queue='%s-tmgr-to-sync'%self._sid,
                                       profiler=local_prof,
                                       logger=self._logger)

                            mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                        except Exception, ex:

                            # Rollback and pass exception
                            self._logger.error('Task %s submission failed, error: %s' % (task.uid, ex))
                            task.state = states.SUBMITTING
                            transition(obj=task,
                                       obj_type='Task',
                                       new_state=states.SUBMITTING,
                                       channel=mq_channel,
                                       queue='%s-tmgr-to-sync'%self._sid,
                                       profiler=local_prof,
                                       logger=self._logger)
                            raise

                except Exception, ex:

                    # Rollback and pass exception
                    logger.error('Error in task execution')
                    raise

                try:

                    # Get request from heartbeat-req for heartbeat response
                    method_frame, props, body = mq_channel.basic_get(queue='%s-heartbeat-req'%self._sid)

                    if body:

                        logger.info('Received heartbeat request')

                        mq_channel.basic_publish(exchange='',
                                                 routing_key='%s-heartbeat-res'%self._sid,
                                                 properties=pika.BasicProperties(correlation_id=props.correlation_id),
                                                 body='response')

                        logger.info('Sent heartbeat response')
                        mq_channel.basic_ack(delivery_tag=method_frame.delivery_tag)

                except Exception, ex:

                    logger.error('Failed to respond to heartbeat request, error: %s' % ex)
                    raise

            local_prof.prof('terminating tmgr process', uid=uid)
            self._mq_connection.close()
            # umgr.unregister_callback(unit_state_cb)
            local_prof.close()

        except KeyboardInterrupt:

            self._logger.error('Execution interrupted by user (you probably hit Ctrl+C), ' +
                               'trying to cancel tmgr process gracefully...')
            raise KeyboardInterrupt

        except Exception, ex:

            print traceback.format_exc()
            raise Error(text=ex)

    # ------------------------------------------------------------------------------------------------------------------
    # Public Methods
    # ------------------------------------------------------------------------------------------------------------------

    def start_heartbeat(self):
        """
        **Purpose**: Method to start the heartbeat thread. The heartbeat function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        if not self._hb_thread:

            try:

                self._logger.info('Starting hearbeat thread')
                self._prof.prof('creating heartbeat thread', uid=self._uid)
                self._hb_thread = threading.Thread(target=self._heartbeat, name='heartbeat')
                self._hb_alive = threading.Event()
                self._prof.prof('starting heartbeat thread', uid=self._uid)
                self._hb_thread.start()

                return True

            except Exception, ex:

                self._logger.error('Heartbeat not started, error: %s' % ex)
                self.end_heartbeat()
                raise

        else:
            self._logger.warn('Heartbeat thread already running, but attempted to restart!')

    def end_heartbeat(self):
        """
        **Purpose**: Method to terminate the heartbeat thread. This method is 
        blocking as it waits for the heartbeat thread to terminate (aka join).

        This is the last method that is executed from the TaskManager and
        hence closes the profiler.
        """

        try:

            if self._hb_thread:

                if self._hb_thread.is_alive():
                    self._hb_alive.set()
                    self._hb_thread.join()

                self._logger.info('Hearbeat thread terminated')

                self._prof.prof('hearbeat thread terminated', uid=self._uid)

                # We close in the hearbeat because it ends after the tmgr process
                self._prof.close()

                return True

        except Exception, ex:
            self._logger.error('Could not terminate hearbeat thread')
            raise

    def start_manager(self):
        """
        **Purpose**: Method to start the tmgr process. The tmgr function
        is not to be accessed directly. The function is started in a separate
        thread using this method.
        """

        if not self._tmgr_process:

            try:

                self._prof.prof('creating tmgr process', uid=self._uid)
                self._tmgr_terminate = Event()

                self._tmgr_process = Process(target=self._tmgr,
                                             name='task-manager',
                                             args=(
                                                 self._uid,
                                                 self._umgr,
                                                 self._rmgr,
                                                 self._logger,
                                                 self._mq_hostname,
                                                 self._port,
                                                 self._pending_queue,
                                                 self._completed_queue)
                                             )

                self._logger.info('Starting task manager process')
                self._prof.prof('starting tmgr process', uid=self._uid)
                self._tmgr_process.start()

                return True

            except Exception, ex:

                self._logger.error('Task manager not started, error: %s' % ex)
                self.end_manager()
                raise

        else:
            self._logger.warn('tmgr process already running, but attempted to restart!')

    def end_manager(self):
        """
        **Purpose**: Method to terminate the tmgr process. This method is 
        blocking as it waits for the tmgr process to terminate (aka join).
        """

        try:

            if self._tmgr_process:

                if not self._tmgr_terminate.is_set():
                    self._tmgr_terminate.set()

                self._tmgr_process.join()
                self._logger.info('Task manager process closed')

                self._prof.prof('tmgr process terminated', uid=self._uid)

                return True

        except Exception, ex:
            self._logger.error('Could not terminate task manager process')
            raise

    def check_alive(self):
        """
        **Purpose**: Check if the tmgr process is alive and running
        """

        return self._tmgr_process.is_alive()

    def check_heartbeat(self):
        """
        **Purpose**: Check if the heartbeat thread is alive and running
        """

        return self._hb_thread.is_alive()

    # ------------------------------------------------------------------------------------------------------------------



import os
import csv
import copy
import glob
import time
import threading

import radical.utils as ru

#from radical.entk.exceptions import *
import traceback
import socket

def read_profiles(profiles):
    """
    We read all profiles as CSV files and parse them.  For each profile,
    we back-calculate global time (epoch) from the synch timestamps.
    """
    ret    = dict()
    fields = ru.Profiler.fields

    for prof in profiles:
        rows = list()
        with open(prof, 'r') as csvfile:
            reader = csv.DictReader(csvfile, fieldnames=fields)
            for row in reader:

                # skip header
                if row['time'].startswith('#'):
                    continue

                row['time'] = float(row['time'])

                # store row in profile
                rows.append(row)
    
        ret[prof] = rows

    return ret

def combine_profiles(profs):
    """
    We merge all profiles and sorted by time.
    This routine expectes all profiles to have a synchronization time stamp.
    Two kinds of sync timestamps are supported: absolute and relative.
    Time syncing is done based on 'sync abs' timestamps, which we expect one to
    be available per host (the first profile entry will contain host
    information).  All timestamps from the same host will be corrected by the
    respectively determined ntp offset.  We define an 'accuracy' measure which
    is the maximum difference of clock correction offsets across all hosts.
    The method returnes the combined profile and accuracy, as tuple.
    """

    # we abuse the profile combination to also derive a pilot-host map, which
    # will tell us on what exact host each pilot has been running.  To do so, we
    # check for the PMGR_ACTIVE advance event in agent_0.prof, and use the NTP
    # sync info to associate a hostname.
    # FIXME: This should be replaced by proper hostname logging in 
    #        in `pilot.resource_details`.

    pd_rel   = dict() # profiles which have relative time refs
    hostmap  = dict() # map pilot IDs to host names

    t_host   = dict() # time offset per host
    p_glob   = list() # global profile
    t_min    = None   # absolute starting point of prof session
    c_qed    = 0      # counter for profile closing tag
    accuracy = 0      # max uncorrected clock deviation

    for pname, prof in profs.iteritems():

        if not len(prof):
          # print 'empty profile %s' % pname
            continue

        if not prof[0]['msg']:
            # FIXME: https://github.com/radical-cybertools/radical.analytics/issues/20 
          # print 'unsynced profile %s' % pname
            continue

        t_prof = prof[0]['time']

        host, ip, t_sys, t_ntp, t_mode = prof[0]['msg'].split(':')
        host_id = '%s:%s' % (host, ip)

        if t_min:
            t_min = min(t_min, t_prof)
        else:
            t_min = t_prof

        if t_mode == 'sys':
          # print 'sys synced profile (%s)' % t_mode
            continue

        # determine the correction for the given host
        t_sys = float(t_sys)
        t_ntp = float(t_ntp)
        t_off = t_sys - t_ntp

        if host_id in t_host:

            accuracy = max(accuracy, t_off-t_host[host_id])

            if abs(t_off-t_host[host_id]) > NTP_DIFF_WARN_LIMIT:
                print 'conflict sync   %-35s (%-35s) %6.1f : %6.1f :  %12.5f' \
                        % (os.path.basename(pname), host_id, t_off, t_host[host_id], (t_off-t_host[host_id]))

            continue # we always use the first match

      # print 'store time sync %-35s (%-35s) %6.1f' \
      #         % (os.path.basename(pname), host_id, t_off)

    unsynced = set()
    for pname, prof in profs.iteritems():

        if not len(prof):
            continue

        if not prof[0]['msg']:
            continue

        host, ip, _, _, _ = prof[0]['msg'].split(':')
        host_id = '%s:%s' % (host, ip)
      # print ' --> pname: %s [%s] : %s' % (pname, host_id, bool(host_id in t_host))
        if host_id in t_host:
            t_off   = t_host[host_id]
        else:
            unsynced.add(host_id)
            t_off = 0.0

        t_0 = prof[0]['time']
        t_0 -= t_min

      # print 'correct %12.2f : %12.2f for %-30s : %-15s' % (t_min, t_off, host, pname) 

        # correct profile timestamps
        for row in prof:

            t_orig = row['time'] 

            row['time'] -= t_min
            row['time'] -= t_off

            # count closing entries
            if row['event'] == 'QED':
                c_qed += 1

            if 'agent_0.prof' in pname    and \
                row['event'] == 'advance' and \
                row['state'] == rps.PMGR_ACTIVE:
                hostmap[row['uid']] = host_id

          # if row['event'] == 'advance' and row['uid'] == os.environ.get('FILTER'):
          #     print "~~~ ", row

        # add profile to global one
        p_glob += prof


      # # Check for proper closure of profiling files
      # if c_qed == 0:
      #     print 'WARNING: profile "%s" not correctly closed.' % prof
      # if c_qed > 1:
      #     print 'WARNING: profile "%s" closed %d times.' % (prof, c_qed)

    # sort by time and return
    p_glob = sorted(p_glob[:], key=lambda k: k['time']) 

    # for event in p_glob:
    #     if event['event'] == 'advance' and event['uid'] == os.environ.get('FILTER'):
    #         print '#=- ', event


    # if unsynced:
    #     # FIXME: https://github.com/radical-cybertools/radical.analytics/issues/20 
    #     # print 'unsynced hosts: %s' % list(unsynced)
    #     pass

    return [p_glob, accuracy, hostmap]


def clean_profile(profile, sid=None):
    """
    This method will prepare a profile for consumption in radical.analytics.  It
    performs the following actions:
      - makes sure all events have a `ename` entry
      - remove all state transitions to `CANCELLED` if a different final state 
        is encountered for the same uid
      - assignes the session uid to all events without uid
      - makes sure that state transitions have an `ename` set to `state`
    """

    entities = dict()  # things which have a uid

    for event in profile:

        uid   = event['uid']
        state = event['state']
        time  = event['time']
        name  = event['event']

        del(event['event'])

        # we derive entity_type from the uid -- but funnel 
        # some cases into the session
        if uid:
            event['entity_type'] = uid.split('.',1)[0]

        elif uid == 'root':
            event['entity_type'] = 'session'
            event['uid']         = sid
            uid = sid

        else:
            event['entity_type'] = 'session'
            event['uid']         = sid
            uid = sid

        if uid not in entities:
            entities[uid] = dict()
            entities[uid]['states'] = dict()
            entities[uid]['events'] = list()

        if name == 'advance':

            # this is a state progression
            assert(state), 'cannot advance w/o state'
            assert(uid),   'cannot advance w/o uid'

            event['event_type'] = 'state'
            skip = False

            if state in rps.FINAL:

                # a final state will cancel any previoud CANCELED state
                if rps.CANCELED in entities[uid]['states']:
                    del (entities[uid]['states'][rps.CANCELED])

                # vice-versa, we will not add CANCELED if a final
                # state already exists:
                if state == rps.CANCELED:
                    if any([s in entities[uid]['states'] 
                        for s in rps.FINAL]):
                        skip = True
                        continue

            if state in entities[uid]['states']:
                # ignore duplicated recordings of state transitions
                skip = True
                continue
              # raise ValueError('double state (%s) for %s' % (state, uid))

            if not skip:
                entities[uid]['states'][state] = event

        else:
            # FIXME: define different event types (we have that somewhere)
            event['event_type'] = 'event'
            entities[uid]['events'].append(event)


    # we have evaluated, cleaned and sorted all events -- now we recreate
    # a clean profile out of them
    ret = list()
    for uid,entity in entities.iteritems():

        ret += entity['events']
        for state,event in entity['states'].iteritems():
            ret.append(event)

    # sort by time and return
    ret = sorted(ret[:], key=lambda k: k['time']) 

    return ret


def get_profiles(src=None):

    if not src:
        src = "%s/%s" % (os.getcwd())


    if os.path.exists(src):
        
        # EnTK profiles are always on localhost
        profiles  = glob.glob("%s/*.prof"   % src)

    else:
        raise Error(text='%s does not exist'%src)

    if len(profiles) == 0:
        raise Error(text='No profiles found at %s'%src)

    try:

        profiles = read_profiles(profiles)
        prof, acc, hostmap = combine_profiles(profiles)

        if not hostmap:
            hostmap = socket.gethostname()

        prof = clean_profile(prof)

        return prof, acc, hostmap

    except Exception as ex:

        # Push the exception raised by child functions
        print traceback.format_exc()
        raise Exception


def get_description():

    pass